import copy
import datetime
import hashlib
import json
import logging
import pathlib
from itertools import combinations
from math import log
from random import sample
from typing import Callable
from typing import List
from typing import Union

import matplotlib.pyplot as plt
import pandas as pd
from pydantic import Field

from phdi.fhir.utils import extract_value_with_resource_path
from phdi.harmonization.utils import compare_strings
from phdi.linkage.mpi import BaseMPIConnectorClient
from phdi.linkage.mpi import DIBBsMPIConnectorClient
from phdi.linkage.utils import datetime_to_str

LINKING_FIELDS_TO_FHIRPATHS = {
    "first_name": "Patient.name.given",
    "last_name": "Patient.name.family",
    "birthdate": "Patient.birthDate",
    "address": "Patient.address.line",
    "zip": "Patient.address.postalCode",
    "city": "Patient.address.city",
    "state": "Patient.address.state",
    "sex": "Patient.gender",
    "mrn": "Patient.identifier.where(type.coding.code='MR').value",
}


def block_data(data: pd.DataFrame, blocks: List) -> dict:
    """
    Generates dictionary of blocked data where each key is a block
    and each value is a distinct list of lists containing the data
    for a given block.

    :param data: A pandas dataframe of records to be linked.
    :param blocks: List of columns to be used in blocks.
    :return: A dictionary of with the keys as the blocks and the
      values as the data within each block, stored as a list of
      lists.
    """
    blocked_data_tuples = tuple(data.groupby(blocks))

    # Convert data to list of lists within dict
    blocked_data = dict()
    for block, df in blocked_data_tuples:
        blocked_data[block] = df.values.tolist()

    return blocked_data


def calculate_log_odds(
    m_probs: dict,
    u_probs: dict,
    file_to_write: Union[pathlib.Path, None] = None,
):
    """
    Calculate the per-field log odds ratio score that two records will
    match in a given field. Measures the likelihood that two records
    match on a column due to being a true match as opposed to random
    chance.

    :param m_probs: A dictionary of m-probabilities computed per field.
    :param u_probs: A dictionary of u_probabilities computed per field.
    :param file_to_write: Optionally, a destination filepath at which
      to write the probabilities in JSON format. Default is None.
    :raises ValueError: If the supplied m- and u- probability dictionaries
      do not share an equal key set.
    """
    if m_probs.keys() != u_probs.keys():
        raise ValueError(
            "m- and u- probability dictionaries must contain the same set of keys"
        )
    log_odds = {}
    for k in m_probs:
        log_odds[k] = log(m_probs[k]) - log(u_probs[k])
    _write_prob_file(log_odds, file_to_write)
    return log_odds


# TODO: We will eventually want to move away from pandas in favor of something
# more light-weight. While pandas is good for pre-computing and model
# examination, it does come with substantial overhead. Maybe make this work
# on a list of lists at some point.
def calculate_m_probs(
    data: pd.DataFrame,
    true_matches: dict,
    cols: Union[List[str], None] = None,
    file_to_write: Union[pathlib.Path, None] = None,
):
    """
    For a given set of patient records, calculate the per-field
    m-probability. The m-probability for field X is defined as the
    probability that a pair of records A and B have the same value in
    X, given that A and B are a true matching pair. This function
    incorporates LaPlacian Smoothing to account for unseen data and
    to resolve future logarithms against 0.

    :param data: A pandas dataframe of patient records to compute
      probabilities for.
    :param true_matches: A dictionary holding the IDs of record pairs
      that are true matches in the data set. The format of the dictionary
      should be such that the IDs of the "lower numbered" records in each
      match pair are the keys, and the values are sets of the "higher
      numbered" records in each pair.
    :param cols: Optionally, a list of columns to compute probabilities
      for. If not supplied, computes probabilities across all fields.
      Default is None.
    :param file_to_write: Optionally, a destination filepath at which to
      write the probabilities in JSON format. Default is None.
    """
    if cols is None:
        cols = data.columns
    m_probs = {c: 1.0 for c in cols}
    total_pairs = 1.0
    for root_record, paired_records in true_matches.items():
        total_pairs += len(paired_records)
        for pr in paired_records:
            for c in cols:
                if data[c].iloc[root_record] == data[c].iloc[pr]:
                    m_probs[c] += 1
    for c in cols:
        m_probs[c] /= total_pairs

    _write_prob_file(m_probs, file_to_write)
    return m_probs


# TODO: We will eventually want to move away from pandas in favor of something
# more light-weight. While pandas is good for pre-computing and model
# examination, it does come with substantial overhead. Maybe make this work
# on a list of lists at some point.
def calculate_u_probs(
    data: pd.DataFrame,
    true_matches: dict,
    n_samples: Union[int, None] = None,
    cols: Union[List, None] = None,
    file_to_write: Union[pathlib.Path, None] = None,
):
    """
    For a given set of patient records, calculate the per-field
    u-probability. The u-probability for field X is defined as the
    probability that a pair of records A and B have the same value in
    X, given that A and B are not a true matching pair. This function
    incorporates LaPlacian Smoothing to account for unseen data and
    to handle future logarithms against 0.

    Note: This function can be slow to compute for large data sets.
    It is recommended to pass only a representative subsample of the
    data to the function (we recommend sampling ~25k candidate pairs
    from a sub-sample of ~25k records), even if the sample operation
    is used.

    :param data: A pandas dataframe of patient records to compute
      probabilities for.
    :param true_matches: A dictionary holding the IDs of record pairs
      that are true matches in the data set. The format of the dictionary
      should be such that the IDs of the "lower numbered" records in each
      match pair are the keys, and the values are sets of the "higher
      numbered" records in each pair.
    :param n_samples: Optionally, a number of samples to take from the
      list of possible pairs to compute probabilities over.
    :param cols: Optionally, a list of columns to compute probabilities
      for. If not supplied, computes probabilities across all fields.
      Default is None.
    :param file_to_write: Optionally, a destination filepath at which to
      write the probabilities in JSON format. Default is None.
    """
    if cols is None:
        cols = data.columns

    u_probs = {c: 1.0 for c in cols}

    # Want only the pairs of candidates that aren't true matches
    base_pairs = list(combinations(data.index, 2))
    neg_pairs = [
        x
        for x in base_pairs
        if x[0] not in true_matches or x[1] not in true_matches[x[0]]
    ]

    if n_samples is not None and n_samples < len(neg_pairs):
        neg_pairs = sample(neg_pairs, n_samples)
    for index in neg_pairs:
        for c in cols:
            if data[c].iloc[index[0]] == data[c].iloc[index[1]]:
                u_probs[c] += 1.0

    for c in cols:
        if n_samples is not None and n_samples < len(neg_pairs):
            u_probs[c] = u_probs[c] / (n_samples + 1.0)
        else:
            u_probs[c] = u_probs[c] / (len(neg_pairs) + 1.0)

    _write_prob_file(u_probs, file_to_write)
    return u_probs


def compile_match_lists(match_lists: List[dict], cluster_mode: bool = False):
    """
    Turns a list of matches of either clusters or candidate pairs found
    during linkage into a single unified structure holding all found matches
    across all rules passes. E.g. if a single pass of a linkage algorithm
    uses three rules, hence generates three dictionaries of matches, this
    function will aggregate the results of those three separate dicts into
    a single unified and deduplicated dictionary. For consistency during
    statistical evaluation, the returned dictionary is always indexed by
    the lower ID of the records in a given pair.

    :param match_lists: A list of the dictionaries obtained during a run
      of the linkage algorithm, one dictionary per rule used in the run.
    :param cluster_mode: An optional boolean indicating whether the linkage
      algorithm was run in cluster mode. Default is False.
    :return: The aggregated dictionary of unified matches.
    """
    matches = {}
    for matches_from_rule in match_lists:
        for matches_within_blocks in matches_from_rule.values():
            for candidate_set in matches_within_blocks:
                # Always index the aggregate by the lowest valued ID
                # for statistical consistency and deduplication
                root_record = min(candidate_set)
                if root_record not in matches:
                    matches[root_record] = set()

                # For clustering, need to add all other records in the cluster
                if cluster_mode:
                    for clustered_record in candidate_set:
                        if clustered_record != root_record:
                            matches[root_record].add(clustered_record)
                else:
                    matched_record = max(candidate_set)
                    matches[root_record].add(matched_record)
    return matches


def eval_perfect_match(feature_comparisons: List, **kwargs) -> bool:
    """
    Determines whether a given set of feature comparisons represent a
    'perfect' match (i.e. whether all features that were compared match
    in whatever criteria was specified for them).

    :param feature_comparisons: A list of 1s and 0s, one for each feature
      that was compared during the match algorithm.
    :return: The evaluation of whether the given features all match.
    """
    return sum(feature_comparisons) == len(feature_comparisons)


def eval_log_odds_cutoff(feature_comparisons: List, **kwargs) -> bool:
    """
    Determines whether a given set of feature comparisons matches enough
    to be the result of a true patient link instead of just random chance.
    This is represented using previously computed log-odds ratios.

    :param feature_comparisons: A list of floats representing the log-odds
      score of each field computed on.
    :return: Whether the feature comparisons score well enough to be
      considered a match.
    """
    if "true_match_threshold" not in kwargs:
        raise KeyError("Cutoff threshold for true matches must be passed.")
    return sum(feature_comparisons) >= kwargs["true_match_threshold"]


def extract_blocking_values_from_record(
    record: dict, blocking_fields: List[dict]
) -> dict:
    """
    Extracts values from a given patient record for eventual use in database
    record linkage blocking. A list of fields to block on, as well as a mapping
    of those fields to any desired transformations of their extracted values,
    is used to fhir-path parse the value out of the incoming patient record.

    Currently supported blocking fields:
    - first_name
    - last_name
    - birthdate
    - address
    - city
    - state
    - zip
    - sex
    - mrn

    Currently supported transformations on extracted fields:
    - first4: the first four characters of the value
    - last4: the last four characters of the value

    :param record: A FHIR-formatted Patient record.
    :param blocking_fields: A List of dictionaries giving the blocking
      fields and any transformations that should be applied to them. Each
      dictionary in the list should include a "value" key with one of the
      supported blocking fields above, and may also optionally contain a
      "transformation" key whose value is one of our supported transforms.
    """

    transform_funcs = {
        "first4": lambda x: x[:4] if len(x) >= 4 else x,
        "last4": lambda x: x[-4:] if len(x) >= 4 else x,
    }

    for block_dict in blocking_fields:
        if "value" not in block_dict:
            raise KeyError(
                f"Input dictionary for block {block_dict} must contain a 'value' key."
            )

    block_vals = dict.fromkeys([b.get("value") for b in blocking_fields], "")
    transform_blocks = [b for b in blocking_fields if "transformation" in b]
    transformations = dict(
        zip(
            [b.get("value") for b in transform_blocks],
            [b.get("transformation") for b in transform_blocks],
        )
    )
    for block_dict in blocking_fields:
        block = block_dict.get("value")
        try:
            # Apply utility extractor for safe parsing
            value = extract_value_with_resource_path(
                record,
                LINKING_FIELDS_TO_FHIRPATHS[block],
                selection_criteria="first",
            )
            if value:
                if block in transformations:
                    try:
                        value = transform_funcs[transformations[block]](value)
                    except KeyError:
                        raise ValueError(
                            f"Transformation {transformations[block]} is not valid."
                        )
                    block_vals[block] = {
                        "value": value,
                        "transformation": transformations[block],
                    }
                else:
                    block_vals[block] = {"value": value}

        except KeyError:
            raise ValueError(f"Field {block} is not a supported extraction field.")

    # Account for any incoming FHIR resources that return no data
    # for a field--don't count this against records to-block
    keys_to_pop = []
    for field in block_vals:
        if _is_empty_extraction_field(block_vals, field):
            keys_to_pop.append(field)
    for k in keys_to_pop:
        block_vals.pop(k)

    return block_vals


def feature_match_exact(
    record_i: List,
    record_j: List,
    feature_col: str,
    col_to_idx: dict[str, int],
    **kwargs: dict,
) -> bool:
    """
    Determines whether a single feature in a given pair of records
    constitutes an exact match (perfect equality).

    :param record_i: One of the records in the candidate pair to evaluate.
    :param record_j: The second record in the candidate pair.
    :param feature_col: The name of the column being evaluated (e.g. "city").
    :param col_to_idx: A dictionary mapping column names to the numeric index
      in which they occur in order in the data.
    :return: A boolean indicating whether the features are an exact match.
    """
    idx = col_to_idx[feature_col]
    return record_i[idx] == record_j[idx]


def feature_match_four_char(
    record_i: List,
    record_j: List,
    feature_col: str,
    col_to_idx: dict[str, int],
    **kwargs: dict,
) -> bool:
    """
    Determines whether a string feature in a pair of records exactly matches
    on the first four characters.

    :param record_i: One of the records in the candidate pair to evaluate.
    :param record_j: The second record in the candidate pair.
    :param feature_col: The name of the column being evaluated (e.g. "city").
    :param col_to_idx: A dictionary mapping column names to the numeric index
      in which they occur in order in the data.
    :return: A boolean indicating whether the features are a match.
    """
    idx = col_to_idx[feature_col]
    first_four_i = record_i[idx][: min(4, len(record_i[idx]))]
    first_four_j = record_j[idx][: min(4, len(record_j[idx]))]
    return first_four_i == first_four_j


def feature_match_fuzzy_string(
    record_i: List,
    record_j: List,
    feature_col: str,
    col_to_idx: dict[str, int],
    **kwargs: dict,
) -> bool:
    """
    Determines whether two strings in a given pair of records are close
    enough to constitute a partial match. The exact nature of the match
    is determined by the specified string comparison function (see
    harmonization/utils/compare_strings for more details) as well as a
    scoring threshold the comparison must meet or exceed.

    :param record_i: One of the records in the candidate pair to evaluate.
    :param record_j: The second record in the candidate pair.
    :param feature_col: The name of the column being evaluated (e.g. "city").
    :param col_to_idx: A dictionary mapping column names to the numeric index
      in which they occur in order in the data.
    :param **kwargs: Optionally, a dictionary including specifications for
      the string comparison metric to use, as well as the cutoff score
      beyond which to classify the strings as a partial match.
    :return: A boolean indicating whether the features are a fuzzy match.
    """
    idx = col_to_idx[feature_col]

    # Convert datetime obj to str using helper function
    if feature_col == "birthdate":
        record_i[idx] = datetime_to_str(record_i[idx])
        record_j[idx] = datetime_to_str(record_j[idx])

    # Special case for two empty strings, since we don't want vacuous
    # equality (or in-) to penalize the score
    if record_i[idx] == "" and record_j[idx] == "":
        return True
    if record_i[idx] is None and record_j[idx] is None:
        return True

    similarity_measure = "JaroWinkler"
    if "similarity_measure" in kwargs:
        similarity_measure = kwargs["similarity_measure"]
    threshold = 0.7
    if "threshold" in kwargs:
        threshold = kwargs["threshold"]
    score = compare_strings(record_i[idx], record_j[idx], similarity_measure)
    return score >= threshold


def feature_match_log_odds_exact(
    record_i: List,
    record_j: List,
    feature_col: str,
    col_to_idx: dict[str, int],
    **kwargs: dict,
) -> float:
    """
    Determines whether two feature values in two records should earn the full
    log-odds similarity score (i.e. they match exactly) or whether they
    should earn no weight (they differ). Used for fields for which fuzzy
    comparisons are inappropriate, such as sex.

    :param record_i: One of the records in the candidate pair to evaluate.
    :param record_j: The second record in the candidate pair.
    :param feature_col: The name of the column being evaluated (e.g. "city").
    :param col_to_idx: A dictionary mapping column names to the numeric index
      in which they occur in order in the data.
    :return: A float of the score the feature comparison earned.
    """
    if "log_odds" not in kwargs:
        raise KeyError("Mapping of columns to m/u log-odds must be provided.")
    col_odds = kwargs["log_odds"][feature_col]
    idx = col_to_idx[feature_col]
    if record_i[idx] == record_j[idx]:
        return col_odds
    else:
        return 0.0


def feature_match_log_odds_fuzzy_compare(
    record_i: List,
    record_j: List,
    feature_col: str,
    col_to_idx: dict[str, int],
    **kwargs: dict,
) -> float:
    """
    Determines the weighted string-odds similarly score earned by two
    feature values in two records, as a function of the pre-computed
    log-odds weights and the string similarity between the two features.
    This scales the full score that would be earned from a perfect
    match to a degree of partial weight appropriate to how similar the
    two strings are.

    :param record_i: One of the records in the candidate pair to evaluate.
    :param record_j: The second record in the candidate pair.
    :param feature_col: The name of the column being evaluated (e.g. "city").
    :param col_to_idx: A dictionary mapping column names to the numeric index
      in which they occur in order in the data.
    :return: A float of the score the feature comparison earned.
    """
    if "log_odds" not in kwargs:
        raise KeyError("Mapping of columns to m/u log-odds must be provided.")
    threshold = 0.7
    if "threshold" in kwargs:
        threshold = kwargs["threshold"]
    col_odds = kwargs["log_odds"][feature_col]
    idx = col_to_idx[feature_col]

    # Convert datetime obj to str using helper function
    if feature_col == "birthdate":
        record_i[idx] = datetime_to_str(record_i[idx])
        record_j[idx] = datetime_to_str(record_j[idx])

    score = compare_strings(record_i[idx], record_j[idx], "JaroWinkler")
    if score < threshold:
        score = 0.0
    return score * col_odds


def generate_hash_str(linking_identifier: str, salt_str: str) -> str:
    """
    Generates a hash for a given string of concatenated patient information. The hash
    serves as a "unique" identifier for the patient.

    :param linking_identifier: The value to be hashed.  For example, the concatenation
      of a patient's name, address, and date of birth, delimited by dashes.
    :param salt_str: The salt to use with the hash. This is intended to prevent
      reverse engineering of the PII used to create the hash.
    :return: The hash of the linking_identifier string.
    """
    hash_obj = hashlib.sha256()
    to_encode = (linking_identifier + salt_str).encode("utf-8")
    hash_obj.update(to_encode)
    return hash_obj.hexdigest()


def link_record_against_mpi(
    record: dict,
    algo_config: List[dict],
    external_person_id: str = None,
    mpi_client: BaseMPIConnectorClient = None,
) -> tuple[bool, str]:
    """
    Runs record linkage on a single incoming record (extracted from a FHIR
    bundle) using an existing database as an MPI. Uses a flexible algorithm
    configuration to allow customization of the exact kind of linkage to
    run. Linkage is assumed to run using cluster membership (i.e. the new
    record must match a certain proportion of existing records all assigned
    to a person in order to match), and if multiple persons are matched,
    the new record is linked to the person with the strongest membership
    percentage.

    :param record: The FHIR-formatted patient resource to try to match to
      other records in the MPI.
    :param algo_config: An algorithm configuration consisting of a list
      of dictionaries describing the algorithm to run. See
      `read_linkage_config` and `write_linkage_config` for more details.
    :returns: A tuple consisting of a boolean indicating whether a match
      was found for the new record in the MPI, followed by the ID of the
      Person entity now associated with the incoming patient (either a
      new Person ID or the ID of an existing matched Person).
    """
    # Initialize MPI client
    if mpi_client is None:
        logging.info("MPI client was None, instatiating new client.")
        mpi_client = DIBBsMPIConnectorClient()

    # Need to bind function names back to their symbolic invocations
    # in context of the module--i.e. turn the string of a function
    # name back into the callable defined in link.py

    algo_config = copy.deepcopy(algo_config)
    logging.info(
        f"Starting _bind_func_names_to_invocations at: {datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
    )
    algo_config = _bind_func_names_to_invocations(algo_config)
    logging.info(
        f"Done with _bind_func_names_to_invocations at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
    )

    # Membership ratios need to persist across linkage passes so that we can
    # find the highest scoring match across all trials
    linkage_scores = {}
    for linkage_pass in algo_config:
        blocking_fields = linkage_pass["blocks"]

        # MPI will be able to find patients if *any* of their names or addresses
        # contains extracted values, so minimally block on the first line
        # if applicable
        logging.info(
            f"Starting extract_blocking_values_from_record at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
        )
        blocking_criteria = extract_blocking_values_from_record(record, blocking_fields)
        logging.info(
            f"Done with extract_blocking_values_from_record at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
        )

        # We don't enforce blocking if an extracted value is empty, so if all
        # values come back blank, skip the pass because the only alt is comparing
        # to all found records
        if len(blocking_criteria) == 0:
            logging.info("No blocking criteria extracted from incoming record.")
            continue
        logging.info(
            f"Starting get_block_data at: {datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
        )
        raw_data_block = mpi_client.get_block_data(blocking_criteria)
        logging.info(
            f"Done with get_block_data at: {datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
        )

        data_block = _convert_given_name_to_first_name(raw_data_block)

        # First row of returned block is column headers
        # Map column name to idx, not including patient/person IDs
        col_to_idx = {v: k for k, v in enumerate(data_block[0][2:])}
        if len(data_block[1:]) > 0:  # Check if data_block is empty
            data_block = data_block[1:]
            logging.info(
                f"Starting _flatten_patient_resource at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
            )
            flattened_record = _flatten_patient_resource(record, col_to_idx)
            logging.info(
                f"Done with _flatten_patient_resource at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
            )

            logging.info(
                f"Starting _group_patient_block_by_person at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
            )
            clusters = _group_patient_block_by_person(data_block)
            logging.info(
                f"Done with _group_patient_block_by_person at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
            )

            # Check if incoming record should belong to one of the person clusters
            kwargs = linkage_pass.get("kwargs", {})
            for person in clusters:
                num_matched_in_cluster = 0.0
                for linked_patient in clusters[person]:
                    logging.info(
                        f"Starting _compare_records at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
                    )
                    is_match = _compare_records(
                        flattened_record,
                        linked_patient,
                        linkage_pass["funcs"],
                        col_to_idx,
                        linkage_pass["matching_rule"],
                        **kwargs,
                    )
                    logging.info(
                        f"Done with _compare_records at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
                    )

                    if is_match:
                        num_matched_in_cluster += 1.0

                # Update membership score for this person cluster so that we can
                # track best possible link across multiple passes
                logging.info(
                    f"Starting to update membership score at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
                )
                belongingness_ratio = num_matched_in_cluster / len(clusters[person])
                if belongingness_ratio >= linkage_pass.get("cluster_ratio", 0):
                    logging.info(
                        f"belongingness_ratio >= linkage_pass.get('cluster_ratio', 0): {datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
                    )
                    if person in linkage_scores:
                        linkage_scores[person] = max(
                            [linkage_scores[person], belongingness_ratio]
                        )
                    else:
                        linkage_scores[person] = belongingness_ratio
                logging.info(
                    f"Done with updating membership score at: {datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
                )
    person_id = None
    matched = False

    # If we found any matches, find the strongest one
    if len(linkage_scores) != 0:
        logging.info(
            f"Starting _find_strongest_link at: {datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
        )
        person_id = _find_strongest_link(linkage_scores)
        matched = True
        logging.info(
            f"Done with _find_strongest_link at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
        )
    logging.info(
        f"Starting mpi_client.insert_matched_patient at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
    )
    person_id = mpi_client.insert_matched_patient(
        record, person_id=person_id, external_person_id=external_person_id
    )
    logging.info(
        f"Done with mpi_client.insert_matched_patient at:{datetime.datetime.now().strftime('%m-%d-%yT%H:%M:%S.%f')}"  # noqa
    )

    return (matched, person_id)


def load_json_probs(path: pathlib.Path):
    """
    Load a dictionary of probabilities from a JSON-formatted file.
    The probabilities correspond to previously computed m-, u-, or
    log-odds probabilities derived from patient records, with one
    score for each field (column) appearing in the data.

    :param path: The file path to load the data from.
    :return: A dictionary of probability scores, one for each field
      in the data set on which they were computed.
    :raises FileNotFoundError: If a file does not exist at the given
      path.
    :raises JSONDecodeError: If the file cannot be read as valid JSON.
    """
    try:
        with open(path, "r") as file:
            prob_dict = json.load(file)
        return prob_dict
    except FileNotFoundError:
        raise FileNotFoundError(f"The specified file does not exist at {path}.")
    except json.decoder.JSONDecodeError as e:
        raise json.decoder.JSONDecodeError(
            "The specified file is not valid JSON.", e.doc, e.pos
        )


def match_within_block(
    block: List[List],
    feature_funcs: dict[str, Callable],
    col_to_idx: dict[str, int],
    match_eval: Callable,
    **kwargs,
) -> List[tuple]:
    """
    Performs matching on all candidate pairs of records within a given block
    of data. Actual partitioning of the data should be done outside this
    function, as it compares all possible pairs within the provided partition.
    Uses a given construction of feature comparison rules as well as a
    match evaluation rule to determine the final verdict on whether two
    records are indeed a match.

    A feature function is of the form "feature_match_X" for some condition
    X; it must accept two records (lists of data), an index i in which the
    feature to compare is stored, and the parameter **kwargs. It must return
    a boolean indicating whether the features "match" for whatever definition
    of match the function uses (i.e. this allows modular logic to apply to
    different features in the compared records). Note that not all features
    in a record need a comparison function defined.

    A match evaluation rule is a function of the form "eval_X" for some
    condition X. It accepts as input a list of booleans, one for each feature
    that was compared with feature funcs, and determines whether the
    comparisons constitute a match according to X.

    :param block: A list of records to check for matches. Each record in
      the list is itself a list of features. The first feature of the
      record must be an "id" for the record.
    :param feature_funcs: A dictionary mapping feature indices to functions
      used to evaluate those features for a match.
    :param col_to_idx: A dictionary mapping column names to the numeric index
      in which they occur in order in the data.
    :param match_eval: A function for determining whether a given set of
      feature comparisons constitutes a match for linkage.
    :return: A list of 2-tuples of the form (i,j), where i,j give the indices
      in the block of data of records deemed to match.
    """
    match_pairs = []

    # Dynamic programming table: order doesn't matter, so only need to
    # check each combo of i,j once
    for i, record_i in enumerate(block):
        for j in range(i + 1, len(block)):
            record_j = block[j]
            feature_comps = [
                feature_funcs[feature_col](
                    record_i, record_j, feature_col, col_to_idx, **kwargs
                )
                for feature_col in feature_funcs
            ]

            # If it's a match, store the result
            is_match = match_eval(feature_comps, **kwargs)
            if is_match:
                match_pairs.append((i, j))

    return match_pairs


# @TODO: Make the data parameter into a list of lists once we finish up
# statistical evaluation--alternatively, allow the function to accept both
# data types, but either way, LoL needs to be in there since that's our
# primary data type to use here.
def perform_linkage_pass(
    data: pd.DataFrame,
    blocks: List,
    feature_funcs: dict[str, Callable],
    matching_rule: Callable,
    cluster_ratio: Union[float, None] = None,
    **kwargs,
) -> dict:
    """
    Performs a partial run of a linkage algorithm using a single rule.
    Each rule in an algorithm is associated with its own pass through the
    data.

    :param data: Currently, a pandas dataframe of records to link. When we
      move out of testing, this should become a LoL.
    :param blocks: A list of column headers to use as blocking assignments
      by which to partition the data.
    :param feature_funcs: A dictionary mapping feature indices to functions
      used to evaluate those features for a match.
    :param matching_rule: A function for determining whether a given set of
      feature comparisons constitutes a match for linkage.
    :param cluster_ratio: An optional parameter indicating, if using the
      algorithm in cluster mode, the required membership percentage a record
      must score with an existing cluster in order to join.
    :return: A dictionary mapping each block found in the pass to the matches
      discovered within that block.
    """
    # Retrieve indices of columns
    cols = list(data.columns.values)
    col_to_idx = dict(zip(cols, range(len(cols))))

    blocked_data = block_data(data, blocks)
    matches = {}
    for block in blocked_data:
        if cluster_ratio:
            matches_in_block = _match_within_block_cluster_ratio(
                blocked_data[block],
                cluster_ratio,
                feature_funcs,
                col_to_idx,
                matching_rule,
                **kwargs,
            )
        else:
            matches_in_block = match_within_block(
                blocked_data[block], feature_funcs, col_to_idx, matching_rule, **kwargs
            )
        matches_in_block = _map_matches_to_record_ids(
            matches_in_block, blocked_data[block], cluster_ratio is not None
        )
        matches[block] = matches_in_block
    return matches


# TODO: Migrate away from pandas eventually
# TODO: If profiling ever gets rolled into the pipeline as a standard
# part of the process, we should revisit that this function is essentially
# "blocking"--the user can't do anything while the plot is shown.
def profile_log_odds(
    data: pd.DataFrame,
    true_matches: dict,
    log_odds: dict,
    exact_cols: List,
    fuzzy_cols: List,
    idx_to_col: dict,
    neg_samples: int = 50000,
) -> None:  # pragma: no cover
    """
    Basic graphical profiler for log-odds histogram analysis. Using the
    raw data and previously known true matches, the function computes one
    list of log-odds scores that that would be earned by true matches under
    a given linkage rule, and another list of scores that would be earned
    by a random sampling of non-matches under the same linkage rule. These
    lists are used to plot bimodal histograms so that the cutoff threshold
    between non-matchces and true matches can be visually determined.

    :param data: A pandas data frame holding the raw patient record data.
    :param true_matches: A dictionary of known true matches in the data.
    :param log_odds: A dictionary whose keys are the column fields of data
      and whose values are the log-odds scores that two values match relative
      to random chance.
    :param exact_cols: A list of columns to be evaluated using equality
      comparisons.
    :param fuzzy_cols: A list of columns to be evaluated using fuzzy weighted
      comparisons.
    :param idx_to_col: A dictionary mapping the number of a column in a list
      representation of the data, to the name of the column in a pandas
      representation.
    :param neg_samples: Optionally, how many non-match samples to compute a
      score for when generating the histogram.
    """
    base_pairs = list(combinations(data.index, 2))
    neg_pairs = [
        x
        for x in base_pairs
        if x[0] not in true_matches or x[1] not in true_matches[x[0]]
    ]
    if neg_samples < len(neg_pairs):
        neg_pairs = sample(neg_pairs, neg_samples)

    data = data.values.tolist()
    cols_to_idx = {}
    for idx in idx_to_col:
        cols_to_idx[idx_to_col[idx]] = idx

    true_match_scores = []
    for root_record, paired_records in true_matches.items():
        for pr in paired_records:
            score = 0.0
            for c in exact_cols:
                score += feature_match_log_odds_exact(
                    data[root_record],
                    data[pr],
                    cols_to_idx[c],
                    idx_to_col=idx_to_col,
                    log_odds=log_odds,
                )
            for c in fuzzy_cols:
                score += feature_match_log_odds_fuzzy_compare(
                    data[root_record],
                    data[pr],
                    cols_to_idx[c],
                    idx_to_col=idx_to_col,
                    log_odds=log_odds,
                )
            true_match_scores.append(score)

    non_match_scores = []
    for record_1, record_2 in neg_pairs:
        score = 0.0
        for c in exact_cols:
            score += feature_match_log_odds_exact(
                data[record_1],
                data[record_2],
                cols_to_idx[c],
                idx_to_col=idx_to_col,
                log_odds=log_odds,
            )
        for c in fuzzy_cols:
            score += feature_match_log_odds_fuzzy_compare(
                data[record_1],
                data[record_2],
                cols_to_idx[c],
                idx_to_col=idx_to_col,
                log_odds=log_odds,
            )
            non_match_scores.append(score)

    min_length = min(len(true_match_scores), len(non_match_scores))
    true_match_scores = true_match_scores[:min_length]
    non_match_scores = non_match_scores[:min_length]

    _, bins, _ = plt.hist(true_match_scores, bins=75, range=[0, 25])
    _ = plt.hist(non_match_scores, bins=bins, alpha=0.5)
    plt.show()


def read_linkage_config(config_file: pathlib.Path) -> List[dict]:
    """
    Reads and generates a record linkage algorithm configuration list from
    the provided filepath, which should point to a JSON file. A record
    linkage configuration list is a list of dictionaries--one for each
    pass in the algorithm it describes--containing information on the
    blocking fields, functions, cluster thresholds, and keyword arguments
    for that pass of the linkage algorithm. For a full example of all the
    components involved in a linkage description structure, see the doc
    string for `write_linkage_config`.

    :param config_file: A `pathlib.Path` string pointing to a JSON file
      that describes the algorithm to decode.
    :return: A list of dictionaries whose values can be passed to the
      various parts of linkage pass function.
    """
    try:
        with open(config_file) as f:
            algo_config = json.load(f)
            # Need to convert function keys back to column indices, since
            # JSON serializes dict keys as strings
            for rl_pass in algo_config.get("algorithm"):
                rl_pass["funcs"] = {
                    int(col): f for (col, f) in rl_pass["funcs"].items()
                }
            return algo_config.get("algorithm", [])
    except FileNotFoundError:
        raise FileNotFoundError(f"No file exists at path {config_file}.")
    except json.decoder.JSONDecodeError as e:
        raise json.decoder.JSONDecodeError(
            "The specified file is not valid JSON.", e.doc, e.pos
        )


def score_linkage_vs_truth(
    found_matches: dict[Union[int, str], set],
    true_matches: dict[Union[int, str], set],
    records_in_dataset: int,
    expand_clusters_pairwise: bool = False,
) -> tuple:
    """
    Compute the statistical qualities of a run of record linkage against
    known true results. This function assumes that matches have already
    been determined by the algorithm, and further assumes that true
    matches have already been identified in the data.

    :param found_matches: A dictionary mapping IDs of records to sets of
      other records which were determined to be a match.
    :param true_matches: A dictionary mapping IDs of records to sets of
      other records which are _known_ to be a true match.
    :param records_in_dataset: The number of records in the original data
      set to-link.
    :param expand_clusters_pairwise: Optionally, whether we need to take
      the cross-product of members within the sets of the match list. This
      parameter only needs to be used if the linkage algorithm was run in
      cluster mode. Default is False.
    :return: A tuple reporting the sensitivity/precision, specificity/recall,
      positive prediction value, and F1 score of the linkage algorithm.
    """

    # If cluster mode was used, only the "master" patient's set will exist
    # Need to expand other permutations for accurate statistics
    if expand_clusters_pairwise:
        new_found_matches = {}
        for root_rec in found_matches:
            if root_rec not in new_found_matches:
                new_found_matches[root_rec] = found_matches[root_rec]
            for paired_record in found_matches[root_rec]:
                if paired_record not in new_found_matches:
                    new_found_matches[paired_record] = set()
                for other_record in found_matches[root_rec]:
                    if other_record > paired_record:
                        new_found_matches[paired_record].add(other_record)
        found_matches = new_found_matches

    # Need division by 2 because ordering is irrelevant, matches are symmetric
    total_possible_matches = (records_in_dataset * (records_in_dataset - 1)) / 2.0
    true_positives = 0.0
    false_positives = 0.0
    false_negatives = 0.0

    for root_record in true_matches:
        if root_record in found_matches:
            true_positives += len(
                true_matches[root_record].intersection(found_matches[root_record])
            )
            false_positives += len(
                found_matches[root_record].difference(true_matches[root_record])
            )
            false_negatives += len(
                true_matches[root_record].difference(found_matches[root_record])
            )
        else:
            false_negatives += len(true_matches[root_record])
    for record in set(set(found_matches.keys()).difference(true_matches.keys())):
        false_positives += len(found_matches[record])

    true_negatives = (
        total_possible_matches - true_positives - false_positives - false_negatives
    )

    print("True Positives:", true_positives)
    print("False Positives:", false_positives)
    print("False Negatives:", false_negatives)

    sensitivity = round(true_positives / (true_positives + false_negatives), 3)
    specificity = round(true_negatives / (true_negatives + false_positives), 3)
    ppv = round(true_positives / (true_positives + false_positives), 3)
    f1 = round(
        (2 * true_positives) / (2 * true_positives + false_negatives + false_positives),
        3,
    )
    return (sensitivity, specificity, ppv, f1)


def write_linkage_config(linkage_algo: List[dict], file_to_write: pathlib.Path) -> None:
    """
    Save a provided algorithm description as a JSON dictionary at the provided
    filepath location. Algorithm descriptions are lists of dictionaries, one
    for each pass of the algorithm, whose keys are parameter values for a
    linkage pass (drawn from the list `"funcs"`, `"blocks"`, `"matching_rule"`,
    and optionally `"cluster_ratio"` and `"kwargs"`) and whose values are
    as follows:

    - `"funcs"` should map to a dictionary mapping column index to the
    name of a function in the DIBBS linkage module (such as
    `feature_match_fuzzy_string`)--note that these are the actual
    functions, not string names of the functions
    - `"blocks"` should map to a list of columns to block on (e.g.
    ["MRN4", "ADDRESS4"])
    - `"matching_rule"` should map to one of the evaluation rule functions
    in the DIBBS linkage module (i.e. `eval_perfect_match`)
    - `"cluster_ratio"` should map to a float, if provided
    - `"kwargs"` should map to a dictionary of keyword arguments and their
    associated values, if provided

    Here's an example of a simple single-pass linkage algorithm that blocks
    on zip code, then matches on exact first name, exact last name, and
    fuzzy date of birth (using, say, Levenshtein similarity with a score
    threshold of 0.8) in dictionary descriptor form (for the sake of the
    example, let's assume the data has the column order first, last, DOB):

    [{
        "funcs": {
            0: feature_match_exact,
            1: feature_match_exact,
            2: feature_match_fuzzy_string,
            3: feature_match_fuzzy_string,
        },
        "blocks": ["ZIP"],
        "matching_rule": eval_perfect_match,
        "kwargs": {
            "similarity-measure": "Levenshtein",
            "threshold": 0.8
        }
    }]

    :param linkage_algo: A list of dictionaries whose key-value pairs correspond
      to the rules above.
    :param file_to_write: The path to the destination JSON file to write.
    """
    algo_json = []
    for rl_pass in linkage_algo:
        pass_json = {}
        pass_json["funcs"] = {col: f.__name__ for (col, f) in rl_pass["funcs"].items()}
        pass_json["blocks"] = rl_pass["blocks"]
        pass_json["matching_rule"] = rl_pass["matching_rule"].__name__
        if rl_pass.get("cluster_ratio", None) is not None:
            pass_json["cluster_ratio"] = rl_pass["cluster_ratio"]
        if rl_pass.get("kwargs", None) is not None:
            pass_json["kwargs"] = {
                kwarg: val for (kwarg, val) in rl_pass.get("kwargs", {}).items()
            }
        algo_json.append(pass_json)
    linkage_json = {"algorithm": algo_json}
    with open(file_to_write, "w") as out:
        out.write(json.dumps(linkage_json))


def _bind_func_names_to_invocations(algo_config: List[dict]):
    """
    Helper method that re-maps the string names of functions to their
    callable invocations as defined within the `link.py` module.
    """
    for lp in algo_config:
        feature_funcs = lp["funcs"]
        for func in feature_funcs:
            if isinstance(feature_funcs[func], str):
                feature_funcs[func] = globals()[feature_funcs[func]]
        if isinstance(lp["matching_rule"], str):
            lp["matching_rule"] = globals()[lp["matching_rule"]]
    return algo_config


def _eval_record_in_cluster(
    block: List[List],
    i: int,
    cluster: set,
    cluster_ratio: float,
    feature_funcs: dict[str, Callable],
    col_to_idx: dict[str, int],
    match_eval: Callable,
    **kwargs,
):
    """
    A helper function used to evaluate whether a given incoming record
    satisfies the matching proportion threshold of an existing cluster,
    and therefore would belong to the cluster.
    """
    record_i = block[i]
    num_matched = 0.0
    for j in cluster:
        record_j = block[j]
        feature_comps = [
            feature_funcs[feature_col](
                record_i, record_j, feature_col, col_to_idx, **kwargs
            )
            for feature_col in feature_funcs
        ]

        is_match = match_eval(feature_comps)
        if is_match:
            num_matched += 1.0
    if (num_matched / len(cluster)) >= cluster_ratio:
        return True
    return False


def _compare_records(
    record: List,
    mpi_patient: List,
    feature_funcs: dict,
    col_to_idx: dict[str, int],
    matching_rule: callable,
    **kwargs,
) -> bool:
    """
    Helper method that compares the flattened form of an incoming new
    patient record to the flattened form of a patient record pulled
    from the MPI.
    """
    # Format is patient_id, person_id, alphabetical list of FHIR keys
    # Don't use the first two ID cols when linking
    feature_comps = [
        _compare_records_field_helper(
            record[2:],
            mpi_patient[2:],
            feature_col,
            col_to_idx,
            feature_funcs,
            **kwargs,
        )
        for feature_col in feature_funcs
    ]
    is_match = matching_rule(feature_comps, **kwargs)
    return is_match


def _compare_records_field_helper(
    record: List,
    mpi_patient: List,
    feature_col: str,
    col_to_idx: dict[str, int],
    feature_funcs: dict,
    **kwargs,
) -> bool:
    if feature_col == "first_name":
        return _compare_name_elements(
            record, mpi_patient, feature_funcs, feature_col, col_to_idx, **kwargs
        )
    elif feature_col in ["address", "city", "state", "zip"]:
        return _compare_address_elements(
            record, mpi_patient, feature_funcs, feature_col, col_to_idx, **kwargs
        )
    else:
        return feature_funcs[feature_col](
            record, mpi_patient, feature_col, col_to_idx, **kwargs
        )


def _compare_address_elements(
    record: List,
    mpi_patient: List,
    feature_funcs: dict,
    feature_col: str,
    col_to_idx: dict[str, int],
    **kwargs,
) -> bool:
    """
    Helper method that compares all elements from the flattened form of an incoming
    new patient record to all elements of the flattened patient record pulled from
    the MPI.
    """
    feature_comp = False
    idx = col_to_idx[feature_col]
    for r in record[idx]:
        feature_comp = feature_funcs[feature_col](
            [r], [mpi_patient[idx]], feature_col, {feature_col: 0}, **kwargs
        )
        if feature_comp:
            break
    return feature_comp


def _compare_name_elements(
    record: List,
    mpi_patient: List,
    feature_funcs: dict,
    feature_col: str,
    col_to_idx: dict[str, int],
    **kwargs,
) -> bool:
    """
    Helper method that compares all elements from the flattened form of an incoming
    new patient record's name(s) to all elements of the flattened
    patient's name(s) pulled from the MPI.
    """
    idx = col_to_idx[feature_col]
    feature_comp = feature_funcs[feature_col](
        [" ".join(record[idx])],
        [mpi_patient[idx]],
        feature_col,
        {feature_col: 0},
        **kwargs,
    )
    return feature_comp


def _condense_extract_address_from_resource(resource: dict, field: str) -> List[str]:
    """
    Formatting function to account for patient resources that have multiple
    associated addresses. Each address is a self-contained object, replete
    with its own `line` property that can hold a list of strings. This
    function condenses that `line` into a single concatenated string, for
    each address object, and returns the result in a properly formatted
    list.

    :param resource: The patient resource to extract the address from.
    :param field: The field to extract the address from.
    :return: A list of strings, each string representing a single address.
    """
    expanded_address_fhirpath = LINKING_FIELDS_TO_FHIRPATHS[field]
    expanded_address_fhirpath = ".".join(expanded_address_fhirpath.split(".")[:-1])
    list_of_address_objects = (
        extract_value_with_resource_path(resource, expanded_address_fhirpath, "all")
        or []
    )
    if not list_of_address_objects:
        return None

    if field == "address":
        list_of_address_lists = [
            ao.get(LINKING_FIELDS_TO_FHIRPATHS[field].split(".")[-1], [])
            for ao in list_of_address_objects
        ]
        list_of_usable_address_elements = [
            " ".join(obj) for obj in list_of_address_lists
        ]
    else:
        list_of_usable_address_elements = []
        for address_object in list_of_address_objects:
            list_of_usable_address_elements.append(
                address_object.get(LINKING_FIELDS_TO_FHIRPATHS[field].split(".")[-1])
            )

    return list_of_usable_address_elements


def _find_strongest_link(linkage_scores: dict) -> str:
    """
    Helper method that determines the highest belongingness level that an
    incoming record achieved against a set of clusers based on existing
    patient records in the MPI. The cluster with the highest belongingness
    ratio is chosen as the Person to link the new record to.
    """
    best_person = max(linkage_scores, key=linkage_scores.get)
    return best_person


def _flatten_patient_resource(resource: dict, col_to_idx: dict) -> List:
    """
    Helper method that flattens an incoming patient resource into a list whose
    elements are the keys of the FHIR dictionary, reformatted and ordered
    according to our "blocking fields extractor" dictionary.
    """
    flattened_record = [
        _flatten_patient_field_helper(resource, f) for f in col_to_idx.keys()
    ]
    flattened_record = [resource["id"], None] + flattened_record
    return flattened_record


def _flatten_patient_field_helper(resource: dict, field: str) -> any:
    """
    Helper function that determines the correct way to flatten a patient's
    FHIR field based on the specific field in question. Names and Addresses,
    because their lists can hold multiple objects, are fetched completely,
    whereas other fields just have their first element used (since historical
    information doesn't matter there).

    For any field for which the value would be `None`, instead use an empty string
    (if the field isn't first_name or address) or a list with one element, the
    empty string (if the field is first_name or address). This ensures that
    future loops over the elements don't disrupt the flow of the matching
    algorithm.
    """
    if field == "first_name":
        vals = extract_value_with_resource_path(
            resource, LINKING_FIELDS_TO_FHIRPATHS[field], selection_criteria="all"
        )
        return vals if vals is not None else [""]
    elif field in ["address", "city", "zip", "state"]:
        vals = _condense_extract_address_from_resource(resource, field)
        return vals if vals is not None else [""]
    else:
        val = extract_value_with_resource_path(
            resource, LINKING_FIELDS_TO_FHIRPATHS[field], selection_criteria="first"
        )
        return val if val is not None else ""


def _group_patient_block_by_person(data_block: List[list]) -> dict[str, List]:
    """
    Helper method that partitions the block of patient data returned from the MPI
    into clusters of records according to their linked Person ID.
    """
    clusters = {}
    for mpi_patient in data_block:
        # Format is patient_id, person_id, alphabetical list of FHIR keys
        if mpi_patient[1] not in clusters:
            clusters[mpi_patient[1]] = []
        clusters[mpi_patient[1]].append(mpi_patient)
    return clusters


def _map_matches_to_record_ids(
    match_list: Union[List[tuple], List[set]], data_block, cluster_mode: bool = False
) -> List[tuple]:
    """
    Helper function to turn a list of tuples of row indices in a block
    of data into a list of tuples of the IDs of the records within
    that block.
    """
    matched_records = []

    # Assumes ID is last column in data set
    if cluster_mode:
        for cluster in match_list:
            new_cluster = set()
            for record_idx in cluster:
                new_cluster.add(data_block[record_idx][-1])
            matched_records.append(new_cluster)
    else:
        for matching_pair in match_list:
            id_i = data_block[matching_pair[0]][-1]
            id_j = data_block[matching_pair[1]][-1]
            matched_records.append((id_i, id_j))
    return matched_records


def _match_within_block_cluster_ratio(
    block: List[List],
    cluster_ratio: float,
    feature_funcs: dict[str, Callable],
    col_to_idx: dict[str, int],
    match_eval: Callable,
    **kwargs,
) -> List[set]:
    """
    A matching function for statistically testing the impact of membership
    ratio to the quality of clusters formed. This function behaves similarly
    to `match_within_block`, except that rather than identifying all pairwise
    candidates which are deemed matches, the function creates a list of
    clusters of patients, where each cluster constitutes what would be a
    single "representative" patient in the database. The formation of
    clusters is determined by the parameter `cluster_ratio`, which defines
    the proportion of other records in an existing cluster that a new
    incoming record must match in order to join the cluster.

    :param block: A list of records to check for matches. Each record in
      the list is itself a list of features. The first feature of the
      record must be an "id" for the record.
    :param cluster_ratio: A float giving the proportion of records in an
      existing cluster that a new incoming record must match in order
      to qualify for membership in the cluster.
    :param feature_funcs: A dictionary mapping feature indices to functions
      used to evaluate those features for a match.
    :param col_to_idx: A dictionary mapping column names to the numeric index
      in which they occur in order in the data.
    :param match_eval: A function for determining whether a given set of
      feature comparisons constitutes a match for linkage.
    :return: A list of 2-tuples of the form (i,j), where i,j give the indices
      in the block of data of records deemed to match.
    """
    clusters = []
    for i in range(len(block)):
        # Base case
        if len(clusters) == 0:
            clusters.append({i})
            continue
        found_master_cluster = False

        # Iterate through clusters to find one that we match with
        for cluster in clusters:
            belongs = _eval_record_in_cluster(
                block,
                i,
                cluster,
                cluster_ratio,
                feature_funcs,
                col_to_idx,
                match_eval,
                **kwargs,
            )
            if belongs:
                found_master_cluster = True
                cluster.add(i)
                break

        # Create a new singleton if no other cluster qualified
        if not found_master_cluster:
            clusters.append({i})
    return clusters


def _is_empty_extraction_field(block_vals: dict, field: str):
    """
    Helper method that determines when a field extracted from an incoming
    record should be considered "empty" for the purpose of blocking.
    Fields whose values are either `None` or the empty string should not
    be used when retrieving blocked records from the MPI, since that
    would impose an artificial constraint (e.g. if an incoming record
    has no `last_name` field, we don't want to retrieve only records
    from the MPI that also have no `last_name`).
    """
    # Means the value extractor found no data in the FHIR resource
    if block_vals[field] == "":
        return True
    # Alternatively, there was "data" there, but it's empty
    elif (
        block_vals[field].get("value") is None
        or block_vals[field].get("value") == ""
        or block_vals[field].get("value") == [""]
    ):
        return True  # pragma: no cover
    return False


def _write_prob_file(prob_dict: dict, file_to_write: Union[pathlib.Path, None]):
    """
    Helper method to write a probability dictionary to a JSON file, if
    a valid path is supplied.

    :param prob_dict: A dictionary mapping column names to the log-probability
      values computed for those columns.
    :param file_to_write: Optionally, a path variable indicating where to
      write the probabilities in a JSON format. Default is None (meaning this
      function would execute nothing.)
    """
    if file_to_write is not None:
        with open(file_to_write, "w") as out:
            out.write(json.dumps(prob_dict))


def add_person_resource(
    person_id: str, patient_id: str, bundle: dict = Field(description="A FHIR bundle")
) -> dict:
    """
    Adds a simplified person resource to a bundle if the patient resource in the bundle
    matches an existing record in the Master Patient Index. Returns the bundle with
    the newly added person resource.

    :param person_id: _description_
    :param patient_id: _description_
    :param bundle: _description_, defaults to Field(description="A FHIR bundle")
    :return: _description_
    """
    person_resource = {
        "fullUrl": f"urn:uuid:{person_id}",
        "resource": {
            "resourceType": "Person",
            "id": f"{person_id}",
            "link": [{"target": {"reference": f"Patient/{patient_id}"}}],
        },
        "request": {
            "method": "PUT",
            "url": f"Person/{person_id}",
        },
    }

    bundle.get("entry", []).append(person_resource)

    return bundle


def _convert_given_name_to_first_name(data: list[list]) -> list[list]:
    """
    In the list of query row results, convert the given_name column (which is a
    list of given names) to a first_name column (which is a space-delimited string
    of given names).

    :param data: List of lists block data.
    :return: List of lists with first_name column.
    """
    result = []
    if not data:
        return result  # empty list, should return an empty list

    if "given_name" not in data[0]:
        return data  # given_name not in data, should return the original

    given_name_idx = data[0].index("given_name")
    for idx, row in enumerate(data):
        val = "first_name" if idx == 0 else " ".join(row[given_name_idx])
        result.append(row[:given_name_idx] + [val] + row[given_name_idx + 1 :])
    return result



from abc import ABC
from abc import abstractmethod


class BaseFhirGeocodeClient(ABC):
    """
    Represents a vendor-agnostic geocoder client designed to process
    FHIR-based data. Implementing classes should define methods to
    geocode from both bundles and resources. Callers should use the
    provided interface functions (e.g., geocode_resource) to interact with
    the underlying vendor-specific client property.
    """

    @abstractmethod
    def geocode_resource(self, resource: dict, overwrite=True) -> dict:
        """
        Performs geocoding, using the implementing client, on the provided resource,
        which is passed in as a dictionary.

        :param resource: A FHIR resource to be geocoded.
        :param overwrite: If true, `resource` is modified in-place;
          if false, a copy of `resource` modified and returned.  Default: `True`
        :return: The geocoded resource as a dict.
        """
        pass  # pragma: no cover

    @abstractmethod
    def geocode_bundle(self, bundle: dict, overwrite=True) -> dict:
        """
        Performs geocoding, using the implementing client, on all supported resources in
        the provided FHIR bundle which is passed in as a dictionary.

        :param bundle: A bundle of FHIR resources.
        :param overwrite: If true, `bundle` is modified in-place;
          if false, a copy of `bundle` modified and returned.  Default: `True`
        :return: The geocoded FHIR bundle as a dict.
        """
        pass  # pragma: no cover

    @staticmethod
    def _store_lat_long_extension(address: dict, lat: float, long: float) -> None:
        """
        Adds extension data for latitude and longitude, if the fields aren't already
        present, to a given FHIR-formatted dictionary holding address fields.
        The latitude and longitude data is added directly to the input dictionary.

        :param address: A FHIR formatted dictionary holding address fields.
        :param lat: The latitude to add to the FHIR data as an extension.
        :param long: The longitude to add to the FHIR data as an extension.
        """
        if "extension" not in address:
            address["extension"] = []

        # Append with a properly resolving URL for FHIR's canonical geospatial
        # structure definition, as all extensions are required to have this
        # attribute; see https://www.hl7.org/fhir/extensibility.html
        address["extension"].append(
            {
                "url": "http://hl7.org/fhir/StructureDefinition/geolocation",
                "extension": [
                    {
                        "url": "latitude",
                        "valueDecimal": lat,
                    },
                    {
                        "url": "longitude",
                        "valueDecimal": long,
                    },
                ],
            }
        )

    @staticmethod
    def _store_census_tract_extension(address: dict, census_tract: str) -> None:
        """
        Adds appropriate extension data for census tract for each element in an address
        line, if the field isn't already present, to a given FHIR-formatted dictionary
        holding address fields. Add the extension data directly to the input dictionary,
        leaving census tract as a FHIR-identified geolocation element.

        :param address: A FHIR formatted dictionary holding address fields
        :param census_tract: The census tract to add to the FHIR data as an extension
        """

        # Append with a properly resolving URL for FHIR's canonical censusTract
        # structure definition, as all extensions are required to have this
        # attribute; see https://www.hl7.org/fhir/extensibility.html

        census_extension = {
            "url": "http://hl7.org/fhir/StructureDefinition/iso21090-ADXP-censusTract",
            "valueString": census_tract,
        }

        if address.get("_line") is None:
            address["_line"] = []
        for element_counter in range(len(address["line"])):
            try:
                address["_line"][element_counter].get("extension").append(
                    census_extension
                )
            except AttributeError:
                address["_line"][element_counter] = {"extension": []}
                address["_line"][element_counter].get("extension").append(
                    census_extension
                )
            except IndexError:
                address["_line"].append({"extension": [census_extension]})



# This script converts patient data from parquet to patient FHIR resources.
import uuid
from datetime import datetime
from typing import Dict
from typing import Tuple


def extract_given_name(data: Dict):
    first_name = data.get("first_name", None)
    middle_name = data.get("middle_name", None)

    given_names = []

    for name in [first_name, middle_name]:
        if name is not None:
            for n in name.split():
                given_names.append(n)

    if len(given_names) > 0:
        return given_names
    else:
        return None


def adjust_birthdate(data: Dict):
    # TODO: remove this function and pass in the `format` parameter to dob
    # standardization in ReadSourceData for LAC
    format = "%d%b%Y:00:00:00.000"
    dob = data.get("birthdate", None)
    if dob is not None and ":" in dob:
        datetime_str = datetime.strptime(dob, format)
        dob = datetime_str.strftime("%Y-%m-%d")
    return dob


def convert_to_patient_fhir_resources(data: Dict) -> Tuple:
    """
    Converts and returns a row of patient data into patient resource in a FHIR-formatted
    patient resouce with a newly generated patient id as well as the
    `external_person_id`.

    :param data: Dictionary of patient data that optionionally includes the following
      fields: mrn, ssn, first_name, middle_name, last_name, home_phone, cell-phone, sex,
      birthdate, address, city, state, zip.
    :return: Tuple of the `external_person_id` and FHIR-formatted patient resource.
    """

    patient_id = str(uuid.uuid4())

    # Iterate through each patient and convert patient data to FHIR resource
    patient_resource = {
        "resourceType": "Patient",
        "id": f"{patient_id}",
        "identifier": [
            {
                "type": {
                    "coding": [
                        {
                            "system": "http://terminology.hl7.org/CodeSystem/v2-0203",
                            "code": "MR",
                        }
                    ]
                },
                "value": f"{data.get('mrn', None)}",
            },
            {
                "type": {
                    "coding": [
                        {
                            "system": "http://terminology.hl7.org/CodeSystem/v2-0203",
                            "code": "SS",
                        }
                    ]
                },
                "value": f"{data.get('ssn', None)}",
            },
        ],
        "name": [
            {
                "family": f"{data.get('last_name', None)}",
                "given": extract_given_name(data),
            }
        ],
        "telecom": [
            {
                "system": "phone",
                "value": f"{data.get('home_phone', None)}",
                "use": "home",
            },
            {
                "system": "phone",
                "value": f"{data.get('cell_phone', None)}",
                "use": "mobile",
            },
            {"value": f"{data.get('email', None)}", "system": "email"},
        ],
        "gender": f"{data.get('sex', None)}",
        "birthDate": adjust_birthdate(data),
        "address": [
            {
                "use": "home",
                "line": [f"{data.get('address', None)}"],
                "city": f"{data.get('city', None)}",
                "state": f"{data.get('state', None)}",
                "postalCode": f"{data.get('zip', None)}",
            }
        ],
    }

    fhir_bundle = {
        "resourceType": "Bundle",
        "type": "batch",
        "id": str(uuid.uuid4()),
        "entry": [
            {
                "fullUrl": f"urn:uuid:{patient_id}",
                "resource": patient_resource,
                "request": {"method": "PUT", "url": f"Patient/{patient_id}"},
            },
        ],
    }

    external_person_id = data.get("person_id", None)
    return (external_person_id, fhir_bundle)



from typing import Union

from smartystreets_python_sdk import ClientBuilder
from smartystreets_python_sdk import StaticCredentials
from smartystreets_python_sdk import us_street
from smartystreets_python_sdk.us_street.lookup import Lookup

from phdi.geospatial.core import BaseGeocodeClient
from phdi.geospatial.core import GeocodeResult


class SmartyGeocodeClient(BaseGeocodeClient):
    """
    Represents a PHDI-supplied geocoding client using the Smarty API.
    Requires an authorization ID as well as an authentication token
    in order to build a street lookup client.
    """

    def __init__(
        self,
        smarty_auth_id: str,
        smarty_auth_token: str,
        licenses: list[str] = ["us-standard-cloud"],
    ):
        self.smarty_auth_id = smarty_auth_id
        self.smarty_auth_token = smarty_auth_token
        creds = StaticCredentials(smarty_auth_id, smarty_auth_token)
        self.__client = (
            ClientBuilder(creds).with_licenses(licenses).build_us_street_api_client()
        )

    @property
    def client(self) -> us_street.Client:
        """
        This property:
          1. defines a private instance variable __client
          2. makes it accessible through the use of .client()

        This property holds a Smarty-specific connection client that
        allows a user to geocode without directly referencing the
        underlying vendor service client.
        """
        return self.__client

    def geocode_from_str(self, address: str) -> Union[GeocodeResult, None]:
        """
        Geocodes the provided address, which is formatted as a string. If the
        result cannot be latitude- or longitude-located, then Smarty failed
        to precisely geocode the address, so no result is returned. Raises
        an error if the provided address is empty.

        :param address: The address to geocode, given as a string.
        :raises ValueError: When the address does not include street number and name.
        :return: A geocoded address (if valid result) or None (if no valid result).
        """

        # The smarty Lookup class will parse a BadRequestError but retry
        # 5 times if the lookup address is blank, so catch that here
        if address == "":
            raise ValueError("Address must include street number and name at a minimum")

        lookup = Lookup(street=address)
        self.__client.send_lookup(lookup)
        return self._parse_smarty_result(lookup)

    def geocode_from_dict(self, address: dict) -> Union[GeocodeResult, None]:
        """
        Geocodes the provided address, which is formatted as a dictionary.

        The given dictionary should conform to standard nomenclature around address
        fields, including:

        * `street`: the number and street address
        * `street2`: additional street level information (if needed)
        * `apartment`: apartment or suite number (if needed)
        * `city`: city to geocode
        * `state`: state to geocode
        * `postal_code`: the postal code to use
        * `urbanization`: urbanization code for area, sector, or regional
        * `development`: (only used for Puerto Rican addresses)

        There is no minimum number of fields that must be specified to use this
        function; however, a minimum of street, city, and state are suggested
        for the best matches.

        :param address: A dictionary with fields outlined above.
        :raises Exception: When the address street is an empty string.
        :return: A geocoded address (if valid result) or None (if no valid result).
        """

        # Smarty geocode requests must include a street level
        # field in the payload, otherwise generates BadRequestError
        if address.get("street", "") == "":
            raise ValueError("Address must include street number and name at a minimum")

        # Configure the lookup with whatever provided address values
        # were in the user-given dictionary
        lookup = Lookup()
        lookup.street = address.get("street", "")
        lookup.street2 = address.get("street2", "")
        lookup.secondary = address.get("apartment", "")
        lookup.city = address.get("city", "")
        lookup.state = address.get("state", "")
        lookup.zipcode = address.get("postal_code", "")
        lookup.urbanization = address.get("urbanization", "")
        lookup.match = "strict"

        self.__client.send_lookup(lookup)
        return self._parse_smarty_result(lookup)

    @staticmethod
    def _parse_smarty_result(lookup) -> Union[GeocodeResult, None]:
        """
        Parses a returned Smarty geocoding result into a GeocodeResult object.
        If the Smarty lookup is null or doesn't include latitude and longitude
        information, returns None instead.

        :param lookup: The us_street.lookup client instantiated for geocoding.
        :return: The geocoded address (if valid result) or None (if no valid result).
        """
        # Valid responses have results with lat/long
        if lookup.result and lookup.result[0].metadata.latitude:
            smartystreets_result = lookup.result[0]
            street_address = [smartystreets_result.delivery_line_1]
            if smartystreets_result.delivery_line_2:
                street_address.append(smartystreets_result.delivery_line_2)

            # Format the Smarty result into our standard dataclass object
            return GeocodeResult(
                line=street_address,
                city=smartystreets_result.components.city_name,
                state=smartystreets_result.components.state_abbreviation,
                postal_code=smartystreets_result.components.zipcode,
                county_fips=smartystreets_result.metadata.county_fips,
                county_name=smartystreets_result.metadata.county_name,
                lat=smartystreets_result.metadata.latitude,
                lng=smartystreets_result.metadata.longitude,
                precision=smartystreets_result.metadata.precision,
            )

        return



from lxml import etree

from .xml_utils import ECR_NAMESPACES
from .xml_utils import get_ecr_message_ids
from .xml_utils import get_xml_element_details
from .xml_utils import validate_xml_attributes
from .xml_utils import validate_xml_elements
from .xml_utils import validate_xml_value

ERROR_MESSAGES = {
    "fatal": [],
    "errors": [],
    "warnings": [],
    "information": [],
    "message_ids": {},
}


def validate_ecr(ecr_message: str, config: dict, include_error_types: list) -> dict:
    """
    Receives an ecr message (a combined RR and eICR), a configuration of
    what fields are to be validated and how they are to be validated, as
    well as a list of error message types to include (default is all fatal
    errors, basic errors, warnings, and information).  The result of
    validation is returned in a dictionary that also contains a boolean
    if the ecr message is valid or not.

    :param ecr_message: A eCR message that contains both eICR and RR fields.
    :param config: A dictionary of the requirements for validation of the
      eCR message.
    :param include_error_types: A list of error message types to include
        in the final validation response.  Default (fatal, errors,
        warnings, information)
    :return: A dictionary containing bool message_valid as well as a
        dictionary containing the validation results/errors.
    """
    _clear_all_errors_and_ids()
    # encoding ecr_message to allow it to be
    #  parsed and organized as an lxml Element Tree Object
    xml = ecr_message.encode("utf-8")
    parser = etree.XMLParser(ns_clean=True, recover=True, encoding="utf-8")

    # we need a try-catch around this to ensure that the ecr message
    # passed in is proper XML - also ensure it's a clinical document
    try:
        parsed_ecr = etree.fromstring(xml, parser=parser)
        parsed_ecr.xpath("//hl7:ClinicalDocument", namespaces=ECR_NAMESPACES)

    except AttributeError:
        _append_error_message(
            error_message_type="fatal", message="eCR Message is not valid XML!"
        )
        return _response_builder(include_error_types=include_error_types)

    _add_message_ids(get_ecr_message_ids(parsed_ecr=parsed_ecr))

    for field in config.get("fields"):
        # extract the specific information from the configuration
        # for the different fields
        cda_path = field.get("cdaPath")

        # get a list of XML elements that match the field configuration
        matched_xml_elements = validate_xml_elements(
            xml_elements=parsed_ecr.xpath(cda_path, namespaces=ECR_NAMESPACES),
            config_field=field,
        )
        error_message_type = (
            field.get("errorType")
            if field.get("errorType") in ERROR_MESSAGES.keys()
            else "errors"
        )
        # if there are no xml elements that were valid for
        # the configuration then store an error for that based
        # upon the configured error message type
        if not matched_xml_elements:
            error_message = "Could not find field. " + get_xml_element_details(
                None, field
            )
            _append_error_message(
                error_message_type=error_message_type, message=error_message
            )
            continue
        # continue the validation steps for xml attributes and values
        attribute_errors = []
        value_errors = []
        for xml_element in matched_xml_elements:
            attribute_errors += validate_xml_attributes(xml_element, field)
            value_errors += validate_xml_value(xml_element, field)
        # this handles a specific case where just ONE xml element
        # must meet the attribute or xml value criteria - otherwise
        # you wil want to include an error.
        if not (
            field.get("validateOne")
            and len(attribute_errors) < len(matched_xml_elements)
            and len(value_errors) < len(matched_xml_elements)
        ):
            for attribute_error in attribute_errors:
                _append_error_message(
                    error_message_type=error_message_type, message=attribute_error
                )
            for value_error in value_errors:
                _append_error_message(
                    error_message_type=error_message_type, message=value_error
                )
    response = _response_builder(include_error_types=include_error_types)
    return response


def _organize_error_messages(include_error_types: list) -> None:
    # utilize the error_types to filter out the different error message
    # types as well as specify the difference between the different error types
    # during the validation process

    # fatal warnings cannot be filtered and will be automatically included!
    # also, let's not wipe out the message_ids dictionary
    for error_type in ERROR_MESSAGES.keys():
        if (
            error_type not in ("fatal", "message_ids")
            and error_type not in include_error_types
        ):
            ERROR_MESSAGES[error_type] = []


def _response_builder(include_error_types: list) -> dict:
    if ERROR_MESSAGES.get("fatal") != []:
        valid = False
    else:
        valid = True
        _append_error_message(
            error_message_type="information",
            message="Validation completed with no fatal errors!",
        )
    _organize_error_messages(include_error_types=include_error_types)

    return {"message_valid": valid, "validation_results": ERROR_MESSAGES}


def _append_error_message(error_message_type: str, message: str) -> None:
    if error_message_type in ERROR_MESSAGES.keys():
        if isinstance(message, list):
            for msg in message:
                if msg is not None and msg.strip() != "":
                    ERROR_MESSAGES[error_message_type].append(msg.strip())
        elif message is not None and message.strip() != "":
            ERROR_MESSAGES[error_message_type].append(message.strip())


def _add_message_ids(ids: dict) -> None:
    if ids:
        ERROR_MESSAGES["message_ids"] = ids
    else:
        ERROR_MESSAGES["message_ids"] = {}


def _clear_all_errors_and_ids() -> None:
    ERROR_MESSAGES["fatal"] = []
    ERROR_MESSAGES["errors"] = []
    ERROR_MESSAGES["information"] = []
    ERROR_MESSAGES["warnings"] = []
    ERROR_MESSAGES["message_ids"] = {}



from phdi.tabulation.tables import load_schema
from phdi.tabulation.tables import validate_schema
from phdi.tabulation.tables import write_data

__all__ = ("load_schema", "validate_schema", write_data)



import logging
import re
from typing import List

import hl7


def standardize_hl7_datetimes(message: str) -> str:
    """
    Prepares an HL7 message for conversion by normalizing / sanitizing
    fields that are known to contain datetime data in problematic formats. This
    function helps messages conform to expectations.

    This function accepts either segments terminated by `\\r` or `\\n`, but always
    returns data with `\\n` as the segment terminator.

    :param message: The raw HL7 message to sanitize.
    :return: The HL7 message with potential problem formats resolved. If the function
      is unable to parse a date, the original value is retained.
    """
    parsed_message: hl7.Message = None
    try:
        # The hl7 module requires \n characters be replaced with \r
        parsed_message = hl7.parse(message.replace("\n", "\r"))

        # MSH-7 - Message date/time
        normalize_hl7_datetime_segment(parsed_message, "MSH", [7])

        # PID-7 - Date of Birth
        # PID-29 - Date of Death
        # PID-33 - Last update date/time
        normalize_hl7_datetime_segment(parsed_message, "PID", [7, 29, 33])

        # PV1-44 - Admission Date
        # PV1-45 - Discharge Date
        normalize_hl7_datetime_segment(parsed_message, "PV1", [44, 45])

        # ORC-9 Date/time of transaction
        # ORC-15 Order effective date/time
        # ORC-27 Filler's expected availability date/time
        normalize_hl7_datetime_segment(parsed_message, "ORC", [9, 15, 27])

        # OBR-7 Observation date/time
        # OBR-8 Observation end date/time
        # OBR-22 Status change date/time
        # OBR-36 Scheduled date/time
        normalize_hl7_datetime_segment(parsed_message, "OBR", [7, 8, 22, 36])

        # OBX-12 Effective date/time of reference range
        # OBX-14 Date/time of observation
        # OBX-19 Date/time of analysis
        normalize_hl7_datetime_segment(parsed_message, "OBX", [12, 14, 19])

        # TQ1-7 Start date/time
        # TQ1-8 End date/time
        normalize_hl7_datetime_segment(parsed_message, "TQ1", [7, 8])

        # SPM-18 Specimen received date/time
        # SPM-19 Specimen expiration date/time
        normalize_hl7_datetime_segment(parsed_message, "SPM", [18, 19])

        # RXA-3 Date/time start of administration
        # RXA-4 Date/time end of administration
        # RXA-16 Substance expiration date
        # RXA-22 System entry date/time
        normalize_hl7_datetime_segment(parsed_message, "RXA", [3, 4, 16, 22])

    # @TODO: Eliminate logging, raise an exception, document the exception
    # in the docstring, and make this fit into our new structure of allowing
    # the caller to implement more robust error handling
    except Exception:
        logging.exception(
            "Exception occurred while cleaning message.  "
            + "Passing through original message."
        )

        return message

    return str(parsed_message).replace("\r", "\n")


def convert_hl7_batch_messages_to_list(
    content: str, delimiter: str = "\n"
) -> List[str]:
    """
    Converts a batch file of messages into a list of strings
    representing parts of the message. This function is based
    on the following header/tail segments:

    * FHS is a "File Header Segment," which is used to head a file
        (group of batches)
    * FTS is a "File Trailer Segment," which defines the end of a file
    * BHS is "Batch Header Segment," which defines the start of a batch
    * BTS is "Batch Trailer Segment," which defines the end of a batch

    The structure of an HL7 Batch looks like this:
    [FHS] (file header segment) { [BHS] (batch header segment)
    { [MSH (zero or more HL7 messages)
    ....
    ....
    ....
    ] }
    [BTS] (batch trailer segment)
    }
    [FTS] (file trailer segment)

    Ignores lines that start with these header/tail segments since we don't want to
    include them in a message.

    :param content: The batch content to turn into a list.
    :param delimiter: The character delimiting messages in the batch.
    :return: A list of separated, cleaned HL7 messages.
    """

    cleaned_batch = _clean_hl7_batch(content)
    message_lines = cleaned_batch.split(delimiter)
    next_message = ""
    output = []

    for line in message_lines:
        if line.startswith(("FHS", "BHS", "BTS", "FTS")):
            continue

        # If we reach a line that starts with MSH and we have
        # content in nextMessage, then by definition we have
        # a full message in next_message and need to append it
        # to output. This will not trigger the first time we
        # see a line with MSH since next_message will be empty
        # at that time.
        if next_message != "" and line.startswith("MSH"):
            output.append(next_message)
            next_message = ""

        # Otherwise, continue to add the line of text to next_message
        if line != "":
            next_message += f"{line}\r"

    # Since our loop only adds messages to output when it finds
    # a line that starts with MSH, the last message would never
    # be added. So we explicitly add it here.
    if next_message != "":
        output.append(next_message)

    return output


def default_hl7_value(
    message: str, segment_id: str, field_num: int, default_value: str
):
    """
    Defaults a field value in an HL7 message.

    This function accepts either segments terminated by `\\r` or `\\n`, but always
    returns data with `\\n` as the segment terminator.

    :param message: A string representing the HL7 message used to modify
      a value.
    :param segment_id: The segment type (MSH, PID, etc) of the field to replace.
    :param field_num: The field number to replace in the segment named by `segment_id`.
    :param default_value: If the selected field is blank, set the
      field's value to this value.
    :return: The HL7 message with default value inserted at the
      specified segment location, if possible. If not, then the
      original message.
    """

    parsed_message: hl7.Message = None
    try:
        # The hl7 module requires \n characters be replaced with \r
        parsed_message = hl7.parse(message.replace("\n", "\r"))

        segment: hl7.Segment = None
        try:
            segment = parsed_message.segment(segment_id=segment_id)
        except KeyError:
            # If the segment is not found, there is nothing to do
            return message

        if (
            segment.extract_field(field_num=field_num) is None
            or segment.extract_field(field_num=field_num) == ""
        ):
            segment.assign_field(value=default_value, field_num=field_num)

    # @TODO: Eliminate logging, raise an exception, document the exception
    # in the docstring, and make this fit into our new structure of allowing
    # the caller to implement more robust error handling
    except Exception:
        logging.exception(
            "Exception occurred while cleaning message.  "
            + "Passing through original message."
        )

        return message

    return str(parsed_message).replace("\r", "\n")


def normalize_hl7_datetime_segment(
    message: hl7.Message, segment_id: str, field_list: list
) -> None:
    """
    Applies datetime normalization to multiple fields in a segment,
    overwriting values in the input segment as necessary.

    :param message: The HL7 message, represented as a list
      of indexable component strings (which is how the HL7 library
      processes and returns messages).
    :param segment_id: The segment type (MSH, PID, etc) of the field to replace.
    :param field_num: The field number to replace in the segment named by `segment_id`.
    :param field_list: The list of field numbers to replace in the segment named
      by `segement_id`.
    """
    try:
        for segment in message.segments(segment_id):
            for field_num in field_list:
                # Datetime value is always in first component
                component_num = 0
                if len(segment) > field_num and segment[field_num][component_num] != "":
                    cleaned_datetime = normalize_hl7_datetime(segment[field_num][0])
                    segment[field_num][0] = cleaned_datetime

    # @TODO: Eliminate logging, raise an exception, document the exception
    # in the docstring, and make this fit into our new structure of allowing
    # the caller to implement more robust error handling
    except KeyError:
        logging.debug(f"Segment {segment_id} not found in message.")


def normalize_hl7_datetime(hl7_datetime: str) -> str:
    """
    Splits HL7 datetime-formatted fields into the following parts:
    <integer 8+ digits>[.<integer 1+ digits>][+/-<integer 4+ digits>]

    Each group of integers is truncated to conform to the HL7
    specification:

    - first integer group: max 14 digits
    - following decimal point: max 4 digits
    - following +/- (timezone): 4 digits

    This normalization facilitates downstream processing using
    cloud providers that have particular requirements for dates.

    :param hl7_datetime: The raw datetime string to clean.
    :return: The datetime string with normalizing substitutions
      performed, or the original HL7 datetime if no matching
      format could be found.
    """

    hl7_datetime_match = re.match(r"(\d{8}\d*)(\.\d+)?([+-]\d+)?", hl7_datetime)

    if not hl7_datetime_match:
        return hl7_datetime

    hl7_datetime_parts = hl7_datetime_match.groups()

    # Start with date base
    normalized_datetime = hl7_datetime_parts[0][:14]  # First 14 digits

    # Add date decimal if present
    if hl7_datetime_parts[1]:
        normalized_datetime += hl7_datetime_parts[1][:5]  # . plus first 4 digits

    # Add timezone information if present
    if hl7_datetime_parts[2] and len(hl7_datetime_parts[2]) >= 5:
        normalized_datetime += hl7_datetime_parts[2][:5]  # +/- plus 4 digits

    return normalized_datetime


def _clean_hl7_batch(batch: str, delimiter: str = "\n") -> str:
    """
    Cleans a batch file by replacing Windows (CR-LF) newlines with
    the specified newline delimiter (LF by default). Also, strips
    vertical tab and file separator characters which can appear
    in input batch file data.

    :param batch: The batch file data to clean.
    :param delimiter: The newline character to standardize.
    :return: The batch file with newline characters standardized
      and file separators removed.
    """
    cleaned_batch = re.sub("[\r\n]+", delimiter, batch)

    # These are unicode for vertical tab and file separator, respectively
    # \u000b appears before every MSH segment, and \u001c appears at the
    # end of the message in some of the data we've been receiving, so
    # we're explicitly removing them here.
    cleaned_batch = re.sub("[\u000b\u001c]", "", cleaned_batch).strip()
    return cleaned_batch



from phdi.fhir.tabulation.tables import drop_invalid
from phdi.fhir.tabulation.tables import extract_data_from_fhir_search
from phdi.fhir.tabulation.tables import extract_data_from_fhir_search_incremental
from phdi.fhir.tabulation.tables import extract_data_from_schema
from phdi.fhir.tabulation.tables import tabulate_data

__all__ = [
    "drop_invalid",
    "extract_data_from_fhir_search",
    "extract_data_from_fhir_search_incremental",
    "extract_data_from_schema",
    "tabulate_data",
]



import datetime
import pathlib
from typing import List
from typing import Literal
from typing import Union

import phonenumbers
import pycountry
from detect_delimiter import detect

from phdi.harmonization.double_metaphone import DoubleMetaphone

FHIR_DATE_FORMAT = "%Y-%m-%d"
FHIR_DATE_DELIM = "-"


def double_metaphone_string(string: str, dmeta=None) -> List[Union[str, None]]:
    """
    Performs the double metaphone phonetic encoding algorithm on the given
    string. Returns a list holding the primary and secondary phonetic
    representations of the string (including None if there is no valid
    secondary encoding). This function expects basic text cleaning (e.g.
    removal of numeric characters, trimming of spaces, etc.) to already
    have been performed.

    :param string: The string to phonetically encode.
    :param dmeta: An optional existing double metaphone object, in the case
      one has already been instantiated for bulk processing.
    :return: A list of the primary and secondary encodings of the given
      string.
    """
    if dmeta is None:
        dmeta = DoubleMetaphone()
    return dmeta(string)


def standardize_country_code(
    raw_country: str, code_type: Literal["alpha_2", "alpha_3", "numeric"] = "alpha_2"
) -> str:
    """
    Identifies the country represented and generates the desired type of the ISO
    3611 standardized country identifier for a given string representation of a country
    (whether a full name such as "United States," or an abbreviation such as "US"
    or "USA"). If the country identifier cannot be determined, returns None.

    Example: If raw_country = "United States of America," then

    * alpha_2 would be "US"
    * alpha_3 would be "USA"
    * numeric would be "840"

    :param raw_country: The string representation of the country to be
      put in ISO 3611 standardized form.
    :param code_type: One of 'alpha_2', 'alpha_3', or 'numeric'; the
      desired identifier type to generate.
    :return: The standardized country identifier found in the resource's addresses.
    """

    # @TODO: Potentially do some minor restructuring around this logic
    # to make it shorter/clearer. A private helper was discussed, but
    # prevailing consensus was that this logic is a guiding nice-to-have
    # here, where it is, so we'll likely keep it. But this todo marks
    # a revisit later around whether the logic can be restructuerd.

    # First, identify what country the input is referencing
    standard = None
    raw_country = raw_country.strip().upper()
    if len(raw_country) == 2:
        standard = pycountry.countries.get(alpha_2=raw_country)
    elif len(raw_country) == 3:
        standard = pycountry.countries.get(alpha_3=raw_country)
        if standard is None:
            standard = pycountry.countries.get(numeric=raw_country)
    elif len(raw_country) >= 4:
        standard = pycountry.countries.get(name=raw_country)
        if standard is None:
            standard = pycountry.countries.get(official_name=raw_country)

    # Then, if we figured that out, convert it to desired form
    if standard is not None:
        if code_type == "alpha_2":
            standard = standard.alpha_2
        elif code_type == "alpha_3":
            standard = standard.alpha_3
        elif code_type == "numeric":
            standard = standard.numeric

    return standard


def standardize_phone(
    raw_phone: Union[str, List[str]], countries: List = [None, "US"]
) -> Union[str, List[str]]:
    """
    Parses phone number and generates its standardized ISO E.164 international format
    for each given phone number and optional list of associated countries. If an input
    phone number can't be parsed, that number returns the empty string. Attempts
    to parse the inputs using the first successful strategy out of the following:

    1. parses the phone number on its own
    2. parses the phone number using the provided list of possible
       associated countries
    3. parses the phone number using the US as country

    :param raw_phone: One or more raw phone number(s) to standardize.
    :param countries: An optional list containing 2 letter ISO codes
      associated with the phone numbers, signifying to which countries
      the phone numbers might belong.
    :return: Either a string or a list of strings, depending on the
      input of raw_phone, holding the standardized phone number(s).
    """

    # Base cases: we always want to try the phone # on its own first;
    # we also want to try the phone # with the US if all else fails
    if None not in countries:
        countries.insert(0, None)
    if "US" not in countries:
        countries.append("US")

    phones_to_clean = raw_phone
    if isinstance(raw_phone, str):
        phones_to_clean = [raw_phone]
    outputs = []

    for phone in phones_to_clean:
        standardized = ""
        for country in countries:
            # We were able to pull the phone # and corresponding country
            try:
                standardized = phonenumbers.parse(phone, country)
                break

            # This combo of given phone # and country isn't valid
            except phonenumbers.phonenumberutil.NumberParseException:
                continue

        # If we got a match, format it according to ISO standards
        if standardized != "" and phonenumbers.is_possible_number(standardized):
            standardized = str(
                phonenumbers.format_number(
                    standardized, phonenumbers.PhoneNumberFormat.E164
                )
            )
            outputs.append(standardized)
        else:
            outputs.append("")

    if isinstance(raw_phone, str):
        return outputs[0]
    return outputs


def standardize_name(
    raw_name: Union[str, List[str]],
    trim: bool = True,
    case: Literal["upper", "lower", "title"] = "upper",
    remove_numbers: bool = True,
) -> Union[str, List[str]]:
    """
    Performs basic standardization (described below) on each given name. Removes
    punctuation characters and performs a variety of additional cleaning operations.
    Other options can be toggled on or off using the relevant parameter.

    All options specified will be applied uniformly to each input name,
    i.e., specifying case = "lower" will make all given names lower case.

    :param raw_name: Either a single string name or a list of strings,
      each representing a name.
    :param trim: If true, strips leading/trailing whitespace;
      if false, retains whitespace. Default: `True`
    :param case: What case to enforce on each name.

      * `upper`: All upper case
      * `lower`: All lower case
      * `title`: Title case

      Default: `upper`
    :remove_numbers: If true, removes numeric characters from inputs;
      if false, retains numeric characters. Default `True`
    :return: Either a string or a list of strings, depending on the
      input of raw_name, holding the cleaned name(s).
    """
    names_to_clean = raw_name
    if isinstance(raw_name, str):
        names_to_clean = [raw_name]
    outputs = []

    for name in names_to_clean:
        # Remove all punctuation
        cleaned_name = "".join([ltr for ltr in name if ltr.isalnum() or ltr == " "])
        if remove_numbers:
            cleaned_name = "".join([ltr for ltr in cleaned_name if not ltr.isnumeric()])
        if trim:
            cleaned_name = cleaned_name.strip()
        if case == "upper":
            cleaned_name = cleaned_name.upper()
        if case == "lower":
            cleaned_name = cleaned_name.lower()
        if case == "title":
            cleaned_name = cleaned_name.title()
        outputs.append(cleaned_name)

    if isinstance(raw_name, str):
        return outputs[0]
    return outputs


def _build_nicknames_db():
    nicknames_to_names = {}
    with open(pathlib.Path(__file__).parent / "phdi_nicknames.csv", "r") as fp:
        for line in fp:
            if line.strip() != "":
                name, nicks = line.strip().split(":", 1)
                for nickname in nicks.split(","):
                    nicknames_to_names[nickname] = name
    return nicknames_to_names


def _validate_date(year: str, month: str, day: str, future: bool = False) -> bool:
    """
    Validates that a date supplied, split out by the different date components
        is a valid date (ie. not 02-30-2000 or 12-32-2000). This function can
        also verify that the date supplied is not greater than now

    :param raw_date: One date in string format to standardize.
    :param existing_format: A python DateTime format used to parse the date
        supplied.  Default: `%Y-%m-%d` (YYYY-MM-DD).
    :param new_format: A python DateTime format used to convert the date
        supplied into.  Default: `%Y-%m-%d` (YYYY-MM-DD).
    :return: A date as a string in the format supplied by new_format.
    """
    is_valid_date = True
    try:
        valid_date = datetime.datetime(int(year), int(month), int(day))
        if future and valid_date > datetime.datetime.now():
            is_valid_date = False
    except ValueError:
        is_valid_date = False

    return is_valid_date


def _standardize_date(
    raw_date: str, date_format: str = FHIR_DATE_FORMAT, future: bool = False
) -> str:
    """
    Validates a date string is a proper date and then standardizes the
    date string into the FHIR Date Standard (YYYY-MM-DD)

    :param raw_date: A date string to standardize.
    param date_format: A python Date format used to parse and order
        the date components from the date string.
        Default: `%Y-%m-%d` (YYYY-MM-DD).
    :param future: A boolean that if True will verify that the date
        supplied is not in the future.
        Default: False
    :return: A date as a string in the FHIR Date Format.
    """
    # get the date delimiter from the date string
    # this is easier than regexp as we won't have to maintain a list
    # of potential delimiters and just look at what the delim is
    # for the date string supplied
    delim = detect(raw_date)
    format_delim = detect(date_format.replace("%", ""))

    # parse out the different date components (year, month, day)
    date_values = raw_date.split(delim)
    format_values = date_format.replace("%", "").lower().split(format_delim)
    date_dict = {}

    # loop through date values and the format values
    #   and create a date dictionary where the format is the key
    #   and the date values are the value ordering the date component values
    #   using the date format supplied
    for format_value, date_value in zip(format_values, date_values):
        date_dict[format_value[0]] = date_value

    # verify that the date components in the date dictionary create a valid
    # date and based upon the future param that the date is not in the future
    if len(date_dict) != 3 or not _validate_date(
        date_dict["y"], date_dict["m"], date_dict["d"], future
    ):
        raise ValueError(f"Invalid date supplied: {raw_date}")

    return (
        date_dict["y"]
        + FHIR_DATE_DELIM
        + date_dict["m"]
        + FHIR_DATE_DELIM
        + date_dict["d"]
    )


def standardize_birth_date(
    raw_dob: str, existing_format: str = FHIR_DATE_FORMAT
) -> str:
    """
    Validates and standardizes a date of birth string into YYYY-MM-DD format.

    :param raw_dob: One date of birth (dob) to standardize.
    :param existing_format: A python DateTime format used to parse the date of
        birth within the Patient resource.  Default: `%Y-%m-%d` (YYYY-MM-DD).
    :return: Date of birth as a string in YYYY-MM-DD format
        or None if date of birth is invalid.
    """

    #  Need to make sure dob is not None or null ("")
    #  or detect() will end up in an infinite loop
    if raw_dob is None or len(raw_dob) == 0:
        raise ValueError("Date of Birth must be supplied!")

    standardized_dob = _standardize_date(
        raw_date=raw_dob, date_format=existing_format, future=True
    )

    return standardized_dob



from phdi.fhir.geospatial.census import CensusFhirGeocodeClient
from phdi.fhir.geospatial.core import BaseFhirGeocodeClient
from phdi.fhir.geospatial.smarty import SmartyFhirGeocodeClient

__all__ = (
    "BaseFhirGeocodeClient",
    "SmartyFhirGeocodeClient",
    "CensusFhirGeocodeClient",
)



import copy
from typing import List
from typing import Literal
from typing import Union

from phdi.harmonization import double_metaphone_string
from phdi.harmonization import DoubleMetaphone
from phdi.harmonization import standardize_birth_date
from phdi.harmonization import standardize_country_code
from phdi.harmonization import standardize_name
from phdi.harmonization import standardize_phone


def double_metaphone_bundle(bundle: dict, overwrite=True) -> dict:
    """
    Performs the double metaphone algorithm on each name of each patient in a
    given FHIR bundle.

    :param bundle: A FHIR bundle of data containing one or more patient
      resources.
    :param overwrite: If true, `data` is modified in-place; if false, a
      copy of `data` modified and returned.  Default: `True`.
    :return: A dictionary mapping the FHIR IDs of patients in the bundle
      to lists holding the double metaphone representations of their
      names for each FHIR use case their resource includes.
    """
    if not overwrite:
        bundle = copy.deepcopy(bundle)

    dmeta = DoubleMetaphone()
    for entry in bundle.get("entry", []):
        resource = entry.get("resource", {})
        if resource.get("resourceType", "") == "Patient":
            double_metaphone_patient(resource, dmeta, overwrite=True)
    return bundle


def double_metaphone_patient(patient: dict, dmeta=None, overwrite=True) -> dict:
    """
    Performs the double metaphone algorithm for each name in a given patient
    resource. The algorithm is performed on each component of the name (first,
    middle, last), and the resulting representations are ordered in a list
    such that the first element is first name, the last element is last name,
    and all other elements are one or more middle names in the order of
    name presentation. These lists of phonetic representations are stored as
    the values of dictionaries whose keys are the FHIR uses of the name in
    the patient resource (e.g. "official"), and all such dictionaries are
    returned to the caller in a list ordered the same as the names within
    the given resource.

    :param patient: A FHIR-formatted JSON dictionary representing a patient
      resource.
    :param dmeta: An optional existing instantiation of a double metaphone
      object for use in bulk processing.
    :param overwrite: If true, `data` is modified in-place; if false, a
      copy of `data` modified and returned.  Default: `True`.
    :return: A list of dictionaries mapping FHIR uses to the phonetic
      representations of names associated with those uses, in presentation
      order (first, middle, last).
    """

    if not overwrite:
        patient = copy.deepcopy(patient)

    for name in patient.get("name", []):
        # Processing last name separately allows us to note in the result
        # whether last name wasn't present ( = [None, None])
        dm_last = double_metaphone_string(name.get("family", ""), dmeta)
        dm_givens = []

        # Each name is processed sequentially because FHIR expects given
        # names to already follow proper presentation order
        for given in name.get("given", []):
            dm_givens.append(double_metaphone_string(given, dmeta))

        # Cleanest way to store computed encodings is as an extension directly
        # within the HumanName objects of the patient's `name` field
        if name.get("extension", []) == []:
            name["extension"] = []

        name.get("extension").append(
            {
                "url": "https://xlinux.nist.gov/dads/HTML/doubleMetaphone.html",
                "extension": [
                    {"url": "familyName", "valueString": dm_last},
                    {"url": "givenName", "valueString": dm_givens},
                ],
            }
        )

    return patient


def standardize_names(
    data: dict,
    trim: bool = True,
    case: Literal["upper", "lower", "title"] = "upper",
    remove_numbers: bool = True,
    overwrite: bool = True,
) -> dict:
    """
    Standardizes all names contained in a given FHIR bundle or a FHIR resource. The
    default standardization behavior is our defined non-numeric, space-trimming, full
    capitalization standardization, but other modes may be specified.

    :param data: A FHIR-formatted JSON dict.
    :param trim: Whether leading/trailing whitespace should be removed. Default: `True`
    :param case: The type of casing that should be used. Default: `upper`
    :param remove_numbers: If true, delete numeric characters; if false leave numbers
      in place. Default: `True`
    :param overwrite: If true, `data` is modified in-place;
      if false, a copy of `data` modified and returned.  Default: `True`
    :return: The bundle or resource with names appropriately standardized.
    """
    # Copy the data if we don't want to overwrite the original
    if not overwrite:
        data = copy.deepcopy(data)

    # Allow users to pass in either a resource or a bundle
    bundle = data
    if "entry" not in data:
        bundle = {"entry": [{"resource": data}]}

    # Handle all resources individually
    for entry in bundle.get("entry"):
        resource = entry.get("resource", {})
        resource = _standardize_names_in_resource(
            resource, trim, case, remove_numbers, overwrite
        )

    if "entry" not in data:
        return bundle.get("entry", [{}])[0].get("resource", {})
    return bundle


def standardize_phones(data: dict, overwrite=True) -> dict:
    """
    Standardizes all phone numbers in a given FHIR bundle or a FHIR resource.
    Standardization is done according to the underlying `standardize_phone`
    function in `phdi.harmonization`.

    :param data: A FHIR bundle or FHIR-formatted JSON dict.
    :param overwrite: If true, `data` is modified in-place;
      if false, a copy of `data` modified and returned.  Default: `True`
    :return: The bundle or resource with phones appropriately standardized.
    """

    if not overwrite:
        data = copy.deepcopy(data)

    # Allow users to pass in either a resource or a bundle
    bundle = data
    if "entry" not in data:
        bundle = {"entry": [{"resource": data}]}

    for entry in bundle.get("entry"):
        resource = entry.get("resource", {})
        resource = _standardize_phones_in_resource(resource, overwrite)

    if "entry" not in data:
        return bundle.get("entry", [{}])[0].get("resource", {})
    return bundle


def _standardize_names_in_resource(
    resource: dict,
    trim: bool = True,
    case: Literal["upper", "lower", "title"] = "upper",
    remove_numbers: bool = True,
    overwrite: bool = True,
) -> dict:
    """
    Standardizes all found names in a given resource.
    The resource can be of any type currently supported by the
    function's logic. At this time, those resources include:

    * Patient

    The parameters to this function match the standardization flags
    used by the underlying `standardize_name` function found in
    `phdi.harmonization`. For more information, see the docstring for
    that function.

    :param resource: A FHIR-formatted JSON dictionary.
    :param trim: Whether leading/trailing whitespace should be removed.
      Default: `True`
    :param case: The type of casing that should be used. Default: `upper`
    :param remove_numbers: Whether to delete numeric characters.
      Default: `True`
    :param overwrite: Whether to replace the original names in the input
      data with the standardized names. Default: `True`
    :return: The resource with appropriately standardized names.
    """

    if not overwrite:
        resource = copy.deepcopy(resource)

    if resource.get("resourceType", "") == "Patient":
        for name in resource.get("name", []):
            # Handle family names
            if "family" in name:
                transformed_name = standardize_name(
                    name.get("family", ""), trim, case, remove_numbers
                )
                name["family"] = transformed_name

            # Given names are stored in a list, as there could be multiple,
            # process them all and take the overall diff for metrics
            if "given" in name:
                transformed_names = [
                    standardize_name(g, trim, case, remove_numbers)
                    for g in name.get("given", [])
                ]
                name["given"] = transformed_names
    return resource


def _standardize_phones_in_resource(
    resource: dict, overwrite=True
) -> Union[dict, None]:
    if not overwrite:
        resource = copy.deepcopy(resource)

    if resource.get("resourceType", "") == "Patient":
        for telecom in resource.get("telecom", []):
            if telecom.get("system") == "phone" and "value" in telecom:
                countries = _extract_countries_from_resource(resource)
                transformed_phone = standardize_phone(
                    telecom.get("value", ""), countries
                )
                telecom["value"] = transformed_phone
    return resource


def _extract_countries_from_resource(
    resource: dict, code_type: Literal["alpha_2", "alpha_3", "numeric"] = "alpha_2"
) -> List[str]:
    """
    Builds a list containing all of the countries, standardized by code_type, in the
    addresses of a given FHIR resource as interpreted by the ISO 3611: standardized
    country identifier. If the resource is not of a supported type, no
    countries will be returned. Currently supported resource types are:

    * Patient

    :param resource: A FHIR-formatted JSON dictionary.
    :param code_type: A string equal to 'alpha_2', 'alpha_3', or 'numeric'
      to specify which type of standard country identifier to generate.
      Default: `alpha_2`
    :return: A list of all the standardized countries found in the resource's
      addresses.
    """
    countries = []
    resource_type = resource.get("resourceType")
    if resource_type == "Patient":
        for address in resource.get("address"):
            country = address.get("country")
            if country:
                countries.append(standardize_country_code(country, code_type))
    return countries


def _standardize_dob_in_resource(
    resource: dict, format: str = "%Y-%m-%d", overwrite=True
) -> Union[dict, None]:
    if not overwrite:
        resource = copy.deepcopy(resource)

    if resource.get("resourceType", "") == "Patient":
        birth_date = resource.get("birthDate")
        transformed_birth_date = standardize_birth_date(birth_date, format)
        resource["birthDate"] = transformed_birth_date
    return resource


def standardize_dob(data: dict, format: str = "%Y-%m-%d", overwrite=True) -> dict:
    """
    Standardizes all birth dates in a given FHIR bundle or a FHIR resource.
    Standardization is done according to the underlying `standardize_dob` function in
    `phdi.harmonization`.  The final birthDate will follow the FHIR STu3/R4 format
    of YYYY-MM-DD which will be stored in the Patient resource.

    :param data: A FHIR bundle or FHIR-formatted JSON dict.
    :param format: A python DateTime format used to parse the birthDate within
      the Patient resource.  Default: `%Y-%m-%d` (also known as YYYY-MM-DD)
    :param overwrite: If true, `data` is modified in-place;
      if false, a copy of `data` modified and returned.  Default: `True`
    :return: The bundle or resource with bith dates appropriately standardized.
    """

    if not overwrite:
        data = copy.deepcopy(data)

    # Allow users to pass in either a resource or a bundle
    bundle = data
    if "entry" not in data:
        bundle = {"entry": [{"resource": data}]}

    for entry in bundle.get("entry"):
        resource = entry.get("resource", {})
        resource = _standardize_dob_in_resource(resource, format, overwrite)

    if "entry" not in data:
        return bundle.get("entry", [{}])[0].get("resource", {})
    return bundle



from phdi.fhir.linkage.link import add_patient_identifier
from phdi.fhir.linkage.link import add_patient_identifier_in_bundle

__all__ = ["add_patient_identifier_in_bundle", "add_patient_identifier"]



import logging
from typing import List
from typing import Literal

import requests

from phdi.cloud.core import BaseCredentialManager
from phdi.transport import http_request_with_retry


def http_request_with_reauth(
    cred_manager: BaseCredentialManager,
    url: str,
    retry_count: int,
    request_type: Literal["GET", "POST"],
    allowed_methods: List[str],
    headers: dict,
    data: dict = None,
) -> requests.Response:
    """
    First, calls :func:`phdi.transport.http.http_request_with_retry`. If the first call
    fails with an authorization error (HTTP status 401), obtains a new token using the
    `cred_manager`. If the original request had an Authorization header, replaces
    it with the new token and re-initiates
    :func:`phdi.transport.http.http_request_with_retry`.

    :param cred_manager: The credential manager used to authenticate to the FHIR server.
    :param url: The url at which to make the HTTP request.
    :param retry_count: The number of times to retry the request, if the
      first attempt fails.
    :param request_type: The type of request to be made.
    :param allowed_methods: The list of allowed HTTP request methods (i.e.,
      POST, PUT, etc.) for the specific URL and query.
    :param headers: JSON-type dictionary of headers to make the request with,
      including Authorization and content-type.
    :param data: JSON data in the case that the request requires data to be
      posted. Default: `None`
    :return: A `requests.Request` object containing the response from the FHIR server.
    """

    response = http_request_with_retry(
        url=url,
        retry_count=retry_count,
        request_type=request_type,
        allowed_methods=allowed_methods,
        headers=headers,
        data=data,
    )

    # Retry with new token in case it expired since creation (or from cache)
    if response.status_code == 401:
        if headers.get("Authorization", "").startswith("Bearer "):
            new_access_token = cred_manager.get_access_token()
            headers["Authorization"] = f"Bearer {new_access_token}"

        response = http_request_with_retry(
            url=url,
            retry_count=retry_count,
            request_type=request_type,
            allowed_methods=allowed_methods,
            headers=headers,
            data=data,
        )

    return response


def upload_bundle_to_fhir_server(
    bundle: dict,
    cred_manager: BaseCredentialManager,
    fhir_url: str,
    max_bundle_size: int = 500,
) -> list[requests.Response]:
    """
    Uploads a FHIR resource bundle to the FHIR server.

    :param bundle: A FHIR bundle (type "batch" or "transaction") to post.  Each entry in
      the bundle must contain a `request` element in addition to a `resource`.
      The FHIR API provides additional details on creating
      [FHIR-conformant batch/transaction](https://hl7.org/fhir/http.html#transaction)
      bundles.
    :param cred_manager: The credential manager used to authenticate to the FHIR server.
    :param fhir_url: The url of the FHIR server to upload to.
    :param max_bundle_size: The maximum number of resources per bundle to upload to
      the FHIR server.
    :return: A `requests.Response` object containing the response from the FHIR server.
    """

    access_token = cred_manager.get_access_token()
    responses = []

    # ensure that bundles are below the set maximum size of resources
    split_bundles = _split_bundle_resources(bundle, max_bundle_size)

    for single_bundle in split_bundles:
        response = http_request_with_reauth(
            cred_manager=cred_manager,
            url=fhir_url,
            retry_count=3,
            request_type="POST",
            allowed_methods=["POST"],
            headers={
                "Authorization": f"Bearer {access_token}",
                "Accept": "application/fhir+json",
                "Content-Type": "application/fhir+json",
            },
            data=single_bundle,
        )
        # FHIR uploads are sent as a batch.  Although the batch succeeds,
        # individual entries within the batch may fail, so we log them here
        if response.status_code == 200:
            response_json = response.json()

            entries = response_json.get("entry", [])
            for entry_index, entry in enumerate(entries):
                entry_response = entry.get("response", {})

                # FHIR bundle.entry.response.status is string type - integer status code
                # plus may inlude a message
                if entry_response and entry_response.get("status", "") not in [
                    "200 OK",
                    "201 Created",
                    "200",
                    "201",
                ]:
                    _log_fhir_server_error(
                        status_code=int(entry_response["status"][0:3]),
                        batch_entry_index=entry_index,
                    )
        else:
            _log_fhir_server_error(response.status_code)
        responses.append(response)

    return responses


def fhir_server_get(url: str, cred_manager: BaseCredentialManager) -> requests.Response:
    """
    Submits a GET request to a FHIR server given a url and access token for
    authentication.

    :param url: A URL specifying a GET request on a FHIR server.
    :param cred_manager: The credential manager used to authenticate to the FHIR server.
    :return: A `requests.Request` object containing the response from the FHIR server.
    """
    access_token = cred_manager.get_access_token()
    # Open connection to the export operation and kickoff process
    response = http_request_with_reauth(
        cred_manager=cred_manager,
        url=url,
        retry_count=3,
        request_type="GET",
        allowed_methods=["GET"],
        headers={"Authorization": f"Bearer {access_token}"},
    )

    _log_fhir_server_error(response.status_code)

    return response


def _log_fhir_server_error(status_code: int, batch_entry_index: int = None) -> None:
    """
    Logs the error for a given an HTTP status code from a FHIR server's response.

    :param status_code: The status code returned by a FHIR server.
    :param batch_entry_index: A zero-based index indicates which
        resource in the FHIR bundle received the error.
        Defaults to None.
    """
    # TODO: We may dedcide to remove logging, and instead report errors back to
    # calling function as raised exceptions.
    batch_decorator = ""
    if batch_entry_index is not None:
        batch_decorator = (
            f"in zero-based message index {batch_entry_index} of FHIR batch "
        )

    if status_code == 401:
        logging.error(
            f"FHIR SERVER ERROR {batch_decorator}- Status Code 401: Failed to "
            + "authenticate."
        )

    elif status_code == 403:
        logging.error(
            f"FHIR SERVER ERROR {batch_decorator}- Status Code 403: User does not "
            + "have permission to make that request."
        )

    elif status_code == 404:
        logging.error(
            f"FHIR SERVER ERROR {batch_decorator}- Status Code 404: Server or "
            + "requested data not found."
        )

    elif status_code == 410:
        logging.error(
            f"FHIR SERVER ERROR {batch_decorator}- Status Code 410: Server has "
            + "deleted this cached data."
        )

    elif str(status_code).startswith(("4", "5")):
        error_message = (
            f"FHIR SERVER ERROR {batch_decorator}- Status code {status_code}"
        )
        logging.error(error_message)


def _split_bundle_resources(bundle: dict, max_bundle_size: int = 500) -> list:
    """
    Receives a FHIR bundle and splits up the resources in the bundle
    if there are more than 500 resources.  Otherwise it just returns the
    orginal bundle unmodified.

    :param bundle: A FHIR bundle containing a number of FHIR resources.
    :param max_bundle_size: The maximum number of resources per bundle to
      determine where the split of large FHIR bundles should occur.
    :return: A list of FHIR bundles; if the bundle has <= 500 resources
    then the list will have only one element.
    """
    resources = bundle.get("entry")
    resource_count = len(resources)
    split_bundles = []

    entry_index = 0
    while entry_index <= resource_count:
        # grab all the resources and place them in the entry list within
        # the new bundle dictionary up to the maximum number specified
        partial_bundle = {"resourceType": "Bundle", "type": "batch", "entry": []}
        partial_bundle["entry"] = resources[entry_index : entry_index + max_bundle_size]
        # add the new split bundle to the list to be returned
        split_bundles.append(partial_bundle)
        entry_index = entry_index + max_bundle_size
    return split_bundles



from functools import lru_cache

from pydantic import BaseSettings


class DBSettings(BaseSettings):
    mpi_db_type: str
    mpi_dbname: str
    mpi_host: str
    mpi_user: str
    mpi_password: str
    mpi_port: str


@lru_cache()
def get_settings() -> dict:
    """
    Load the values specified in the Settings class from the environment and return a
    dictionary containing them. The dictionary is cached to reduce overhead accessing
    these values.

    :return: A dictionary with keys specified by the Settings. The value of each key is
    read from the corresponding environment variable.
    """
    return DBSettings().dict()



from abc import ABC
from abc import abstractmethod
from typing import List
from typing import Union


class BaseCredentialManager(ABC):
    """
    Provides a common interface for managing service credentials.
    """

    @abstractmethod
    def get_credential_object(self) -> object:
        """
        Gets a cloud-specific credential object.

        :return: A credential object.
        """
        pass  # pragma: no cover

    @abstractmethod
    def get_access_token(self) -> str:
        """
        Gets an access token using the managed credentials.

        :return: An access token.
        """
        pass  # pragma: no cover


class BaseCloudStorageConnection(ABC):
    @abstractmethod
    def download_object(
        self, container_name: str, filename: str, encoding: str = "utf-8"
    ) -> str:
        """
        Downloads a blob from storage.

        :param container_name: The name of the container containing object to download.
        :param filename: The location of file within storage.
        :param encoding: The character encoding applied to the downloaded content.
        :return: The `stream` parameter, if supplied. Otherwise a new stream object
          containing blob content.
        """
        pass  # pragma: no cover

    @abstractmethod
    def upload_object(
        self,
        message: Union[str, dict],
        container_name: str,
        filename: str,
    ) -> None:
        """
        Uploads the content of a given message to blob storage.
        The message can be passed either as a raw string or as JSON.

        :param message: The contents of a message, encoded either as a
          string or in a JSON format.
        :param container_name: The name of the target container for upload.
        :param filename: The location of file within storage container.
        """
        pass  # pragma: no cover

    @abstractmethod
    def list_containers(self) -> List[str]:
        """
        Lists names for this CloudContainerConnection's containers.

        :return: A list of container names.
        """
        pass  # pragma: no cover

    @abstractmethod
    def list_objects(self, container_name: str, prefix: str) -> List[str]:
        """
        Lists names for objects within a container.

        :param container_name: The name of the container to look for objects.
        :param prefix: Filter the objects returned to filenames beginning
          with this value.
        :return: A list of objects within a container.
        """
        pass  # pragma: no cover



import json
import pathlib
import urllib.parse
import warnings
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from urllib.parse import parse_qs
from urllib.parse import urlencode

import requests

from phdi.cloud.core import BaseCredentialManager
from phdi.fhir.transport import http_request_with_reauth
from phdi.fhir.utils import extract_value_with_resource_path
from phdi.tabulation.tables import load_schema
from phdi.tabulation.tables import write_data


def drop_invalid(data: List[list], schema: Dict, table_name: str) -> List[list]:
    """
    Removes resources from tabulated data if the resource contains an invalid value, as
    specified in the invalid_values field in a user-defined schema. Users may provide
    invalid values as a list, including empty string values ("") and
    None/null values (null).
    :param data: A list of lists containing data for a table. The first list in
        the data value is a list of headers serving as the columns, and all subsequent
        lists are rows in the table.
    :param schema: A declarative, user-defined specification, for one or more tables,
        that defines the metadata, properties, and columns of those tables as they
        relate to FHIR resources.
    :param table_name: Name of the table to drop invalid values.
    :param return: A list of lists, without rows of data derived from the FHIR
        resources and elements that contained invalid values.
        The first list in the data value is a list of headers serving as the
        columns, and all subsequent lists are rows in the table.
    """
    invalid_values_by_column_index = {}

    # Identify columns to drop invalid values for each table in schema
    columns = schema["tables"][table_name]["columns"]
    # Identify indices in List of Lists to check for invalid values
    invalid_values_by_column_index[table_name] = {
        i: columns[col].get("invalid_values")
        for i, col in enumerate(columns)
        if columns[col].get("invalid_values", [])
    }

    # Check if resource contains invalid values to be dropped
    rows_to_remove = []
    if len(invalid_values_by_column_index) > 0:
        for i in range(len(data)):
            for index, invalid_values in invalid_values_by_column_index[
                table_name
            ].items():
                if data[i][index] in invalid_values:
                    rows_to_remove.append(i)
                    break

    # Remove rows with invalid values
    for idx, i in enumerate(rows_to_remove):
        del data[i - idx]

    return data


def extract_data_from_fhir_search(
    search_url: str, cred_manager: BaseCredentialManager = None
) -> List[dict]:
    """
    Performs a FHIR search, continuously using the "next" url to perform
    search continuations until no additional search results are available.
    Returns a dictionary containing the data from all search responses.
    :param search_url: The URL to a FHIR server with search criteria.
    :param cred_manager: The credential manager used to authenticate to the FHIR server.
    :raises KeyError: If the query returns no data from the FHIR server.
    :return: A list of FHIR resources returned from the search.
    """

    results, next = extract_data_from_fhir_search_incremental(
        search_url=search_url, cred_manager=cred_manager
    )

    while next is not None:
        incremental_results, next = extract_data_from_fhir_search_incremental(
            search_url=next, cred_manager=cred_manager
        )
        results.extend(incremental_results)

    # Check that results are not empty
    if not results:
        raise ValueError(
            f"No data returned from server with the following query: {search_url}"
        )

    return results


def extract_data_from_fhir_search_incremental(
    search_url: str, cred_manager: BaseCredentialManager = None
) -> Tuple[List[dict], str]:
    """
    Performs a FHIR search for a single page of data and returns a dictionary containing
    the data and a next URL. If there is no next URL (this is the last page of data),
    then return None as the next URL.
    :param search_url: The URL to a FHIR server with search criteria.
    :param cred_manager: The credential manager used to authenticate to the FHIR server.
    :raises requests.HttpError: If the HTTP request was unsuccessful.
    :return: Tuple containing single page of data as a list of dictionaries and the next
        URL.
    """

    # TODO: Modify fhir_server_get (and http_request_with_reauth) to function without
    # mandating a credential manager. Then replace the direct call to
    # http_request_with_reauth with fhir_server_get.
    # response = fhir_server_get(url=full_url, cred_manager=cred_manager)
    headers = {}
    if cred_manager is not None:
        access_token = cred_manager.get_access_token()
        headers = {
            "Authorization": f"Bearer {access_token}",
            "Accept": "application/fhir+json",
            "Content-Type": "application/fhir+json",
        }

    response = http_request_with_reauth(
        url=search_url,
        cred_manager=cred_manager,
        retry_count=2,
        request_type="GET",
        allowed_methods=["GET"],
        headers=headers,
    )

    if response.status_code != 200:  # pragma: no cover
        raise requests.HTTPError(response=response)

    next_url = None
    content = json.loads(response._content.decode("utf-8"))
    if len(content) == 0:
        warnings.warn(
            message=f"The search_url returned no incremental results: {search_url}",
        )
        content = []
    else:
        for link in content.get("link", []):
            if link.get("relation") == "next":
                next_url = link.get("url")

        content = content.get("entry")

    return content, next_url


def extract_data_from_schema(
    schema: dict, fhir_url: str, cred_manager: BaseCredentialManager = None
) -> Dict[str, List[dict]]:
    """
    Performs a full FHIR search for each table in the specified `schema`,
    and returns a dictionary mapping the table name to corresponding search results.
    :param schema: A declarative, user-defined specification, for one or more tables,
        that defines the metadata, properties, and columns of those tables as they
        relate to FHIR resources.
    :param cred_manager: The credential manager used to authenticate to the FHIR server.
    :return: A dict containing the mapping of a table and its columns, grouped by
        table name, to a list of FHIR resource element results returned from
        the search for each subsequent table name.
    """

    search_urls = _generate_search_urls(schema=schema)

    results = {}
    for table_name, search_url in search_urls.items():
        results[table_name] = extract_data_from_fhir_search(
            search_url=f"{fhir_url}/{search_url}", cred_manager=cred_manager
        )

    return results


def tabulate_data(data: List[dict], schema: dict, table_name: str) -> List[list]:
    """
    Transforms a list of FHIR bundle resource entries into a tabular
    format (given by a list of lists) using a user-defined schema.
    Tabulation works using a two-pass procedure. First, resources
    that are associated with one another in the provided schema
    (identified by the presence of a `reference_location` field in
    one of the schema's columns) are grouped together. For each
    table, one type of resource serves as the "anchor", which
    defines the number of rows in the table, while referenced
    resources are either "forwards" or "reverse" references,
    depending on their relationship to the anchor type. Second,
    the aggregated resources are parsed for value extraction using
    the schema's columns, and the results are stored in a list of
    lists for that table. The first entry in this list are the headers
    of the data, taken from the schema. This functions performs the
    above procedure on one table from the schema, specified by a
    table name.
    :param data: A list of FHIR bundle resource entries to tabulate.
    :param schema: A declarative, user-defined specification, for one or more tables,
        that defines the metadata, properties, and columns of those tables as they
        relate to FHIR resources.
    :param table_name: A string specifying the name of a table defined
      in the given schema.
    :raises KeyError: If the given `table_name` does not occur in the
      provided schema.
    :return: A list of lists denoting the tabulated form of the data.
      The first list is a list of headers serving as the columns,
      and all subsequent lists are rows in the table.
    """

    if table_name not in schema.get("tables", {}):
        raise KeyError(f"Provided table name {table_name} not found in schema")

    # First pass: build mapping of references for easy lookup
    ref_directions = _get_reference_directions(schema)
    ref_dicts = _build_reference_dicts(data, ref_directions)

    # Get the columns from the schema so we always iterate through
    # them in a consistent order
    table_params = schema["tables"][table_name]
    column_items = table_params["columns"].items()
    headers = [column_name for column_name, _ in column_items]
    tabulated_data = [headers]
    anchor_type = schema["tables"][table_name]["resource_type"]

    # Second pass over just the anchor data, since that
    # defines the table's rows
    for anchor_resource, is_result_because in (
        ref_dicts.get(table_name, {}).get(anchor_type, {}).values()
    ):
        # Resources that aren't matches to the original criteria
        # don't generate rows because they were included via a
        # reference
        if is_result_because != "match":
            continue

        row = []

        for _, column_params in column_items:
            path_to_use = column_params["fhir_path"]
            resource_to_use = anchor_resource

            # Determine if we need to make a lookup in our
            # first-pass reference mapping
            if "reference_location" in column_params:
                resource_to_use = _dereference_included_resource(
                    resource_to_use,
                    path_to_use,
                    anchor_resource,
                    column_params,
                    ref_dicts,
                    table_name,
                )
                if resource_to_use is None:
                    row.append(None)
                    continue

            # Forward pointers are many-to-one anchor:target (i.e. many patients
            # could point to the same general practitioner), so we only need a
            # single value for them
            if isinstance(resource_to_use, dict):
                row.append(
                    extract_value_with_resource_path(
                        resource_to_use,
                        path_to_use,
                        column_params["selection_criteria"],
                    )
                )

            # Reverse pointers are one-to-many (one patient could have multiple
            # observations pointing to them), so they need to be stored in a list
            else:
                values = [
                    extract_value_with_resource_path(
                        r, path_to_use, column_params["selection_criteria"]
                    )
                    for r in resource_to_use
                ]
                row.append(values)

        tabulated_data.append(row)

    # Drop invalid values specified in the schema
    tabulated_data = drop_invalid(tabulated_data, schema, table_name)

    return tabulated_data


def _build_reference_dicts(data: List[dict], directions_by_table: dict) -> dict:
    """
    Groups resources previously determined to reference each other into
    dictionaries accessed using resource IDs. For each table, a dictionary
    is created whose keys are the ID of the anchor resource that the
    referenced resources relate to, and whose values are pointers to the
    referenced resources. This allows the `tabulate_data` function to
    simply iterate through the anchor resources (which are rows in the
    table) and use its ID to quickly fetch all related resources for
    columnar value extraction.
    :param data: A list of FHIR bundle resource entries to tabulate.
    :param directions_by_table: The output of the `_get_reference_directions`
      function, which provides the directionality of linked resources to
      the anchors they reference.
    :return: A dict holding, for each table, the groups of resources
      from which column values will be extracted.
    """

    # Build up connections table by table, since one resource could be
    # used in multiple different tables
    reference_dicts = {}
    for table_name in directions_by_table.keys():
        reference_dicts[table_name] = {}

    for entry in data:
        resource = entry.get("resource", {})
        current_resource_type = resource.get("resourceType", "")

        # Check each resource we got back against each table's schema
        # to see if it slots in as an anchor, a forward reference, or
        # a reverse reference
        for table_name, resource_directions in directions_by_table.items():
            if (
                current_resource_type == resource_directions["anchor"]
                or current_resource_type in resource_directions["forward"]
            ):
                if current_resource_type not in reference_dicts[table_name]:
                    reference_dicts[table_name][current_resource_type] = {}

                # Forward pointers are easy: just use the resource's ID, since
                # that's what the anchor will reference; store as a tuple since
                # it's possible for an anchor resource to reference another
                # resource of the same type without the reference needing
                # to generate a row
                reference_dicts[table_name][current_resource_type][
                    resource.get("id", "")
                ] = (resource, entry.get("search", {}).get("mode", ""))

            if current_resource_type in resource_directions["reverse"]:
                if current_resource_type not in reference_dicts[table_name]:
                    reference_dicts[table_name][current_resource_type] = {}

                # Reverse pointers are more involved: need to figure out what
                # resource this points to
                ref_loc = directions_by_table[table_name]["reverse"][
                    current_resource_type
                ]
                ref_path = ref_loc.replace(":", ".") + ".reference"
                referenced_anchor = extract_value_with_resource_path(resource, ref_path)
                referenced_anchor = referenced_anchor.split("/")[-1]

                # There could be a many-to-one relationship with reverse pointers,
                # so store them in a list
                if (
                    referenced_anchor is not None
                    and referenced_anchor
                    not in reference_dicts[table_name][current_resource_type]
                ):
                    reference_dicts[table_name][current_resource_type][
                        referenced_anchor
                    ] = []
                reference_dicts[table_name][current_resource_type][
                    referenced_anchor
                ].append(resource)

    return reference_dicts


def _dereference_included_resource(
    resource_to_use: dict,
    path_to_use: str,
    anchor_resource: dict,
    column_params: dict,
    ref_dicts: dict,
    table_name: str,
) -> Union[dict, None]:
    anchor_id = anchor_resource.get("id", "")
    [direction, ref_path] = column_params["reference_location"].split(":", 1)
    referenced_type = path_to_use.split(".")[0]

    # If a reference resource is requested but the extracted
    # data didn't contain any of them, the referenced type
    # doesn't appear in the mapping
    if referenced_type not in ref_dicts[table_name]:
        return None  # pragma: no cover

    # An anchor resource references another resource, so get the
    # ID from the anchor and look it up
    if direction == "forward":
        path_to_reference = ref_path.replace(":", ".") + ".reference"
        referenced_id = extract_value_with_resource_path(
            anchor_resource, path_to_reference
        )

        # The requested resource may not exist
        if referenced_id not in ref_dicts[table_name][referenced_type]:
            return None
        resource_to_use = ref_dicts[table_name][referenced_type][referenced_id][0]

    # Another resource references our anchor resource
    else:
        if anchor_id not in ref_dicts[table_name][referenced_type]:
            return None
        resource_to_use = ref_dicts[table_name][referenced_type][anchor_id]

    return resource_to_use


def _generate_search_url(
    url_with_querystring: str, default_count: int = None, default_since: str = None
) -> str:
    """
    Generates a FHIR query string using the supplied search string, defaulting values
    for `_count` and `_since`, if given and not already set in the
    `url_with_querystring`.
    :param url_with_querystring: The search URL with querystring. The search URL
      may contain the base URL, or may start with the resource name.
    :param default_count: If set, and querystring does not specify `_count`, the
      `_count` parameter is added to the query string with this value. Default: `None`
    :param default_since: If set, and querystring does not specify `_since`, the
      `_since` parameter is added to the query string with this value. Default: `None`
    :return: The `url_with_querystring` including any defaulted values.
    """
    if "?" in url_with_querystring:
        search_url_prefix, search_query_string = url_with_querystring.split("?", 1)
    else:
        # Split will generate a ValueError if the delimiter is not found
        # in the string, so handle this as an edge case.
        search_url_prefix, search_query_string = (url_with_querystring, "")

    query_string_dict = parse_qs(search_query_string)
    if default_count is not None and query_string_dict.get("_count") is None:
        query_string_dict["_count"] = [default_count]

    if default_since is not None and query_string_dict.get("_since") is None:
        query_string_dict["_since"] = [default_since]

    updated_query_string = urlencode(query_string_dict, doseq=True)
    if not updated_query_string:
        return search_url_prefix

    return "?".join((search_url_prefix, urlencode(query_string_dict, doseq=True)))


def _merge_include_query_params_for_location(
    query_params: dict, reference_location: str
) -> dict:
    """
    Merges an _include and/or _revinclude search parameter into the supplied
    query parameters based on the supplied reference location.
    :param query_params: A dictionary containing query parameters of the form
      `{ "param_name": "param_value" }` or
      `{ "param_name": ["param_value1", ...]}`.
    :param reference_location: The FHIR resource type and field location for
      the referenced resource. For more informaiton see the
      [FHIR documentation](https://www.hl7.org/fhir/search.html#revinclude).
    :return: The modified `query_params` input parameter. Since the
      `query_params` dict is modified in place, the caller can access the
      result in the original input parameter if called with a variable
      or the return value.
    """

    if reference_location == "":
        raise ValueError("reference_location cannot be empty")

    direction, field_location = reference_location.split(":", 1)
    resource, element_path = field_location.split(":", 1)

    # Convert field location to search parameter formatting
    new_field_location = ""
    for letter in element_path:
        if letter.isupper():
            new_field_location += "-"
        new_field_location += letter.lower()

    field_location = resource + ":" + new_field_location

    # Search term is _include for forward searchs, _revinclude for reverse searches.
    # In addition, we must add an :iterate modifier if the reference is relative to
    # another included resource
    query_param_direction = None
    if direction == "forward":
        query_param_direction = "_include"
    elif direction == "reverse":
        query_param_direction = "_revinclude"

    referenced_resource_types = query_params.get(query_param_direction)

    # Handle the case where the search term (_include or _revinclude)
    # is not specified or is specified as a list.
    if referenced_resource_types is None:
        referenced_resource_types = []
        query_params[query_param_direction] = referenced_resource_types
    elif isinstance(referenced_resource_types, str):
        # Convert current_referenced_direction from str to list, and
        # make sure the query_params dict references the new object.
        referenced_resource_types = [referenced_resource_types]
        query_params[query_param_direction] = referenced_resource_types

    if field_location not in referenced_resource_types:
        referenced_resource_types.append(field_location)

    return query_params


def _generate_search_urls(schema: dict) -> dict:
    """
    Parses a schema, and populates a dictionary containing generated search strings
    for each table, in the following structure:
    * table_1: search_string_1
    * table_2: search_string_2
    * ...
    :param schema:  A declarative, user-defined specification, for one or more tables,
        that defines the metadata, properties, and columns of those tables as they
        relate to FHIR resources.
    :raises ValueError: If any table does not contain a `search_string` entry.
    :return: A dict containing search URLs.
    """
    url_dict = {}

    schema_metadata = schema.get("metadata", {})
    count_top = schema_metadata.get("results_per_page")
    since_top = schema_metadata.get("earliest_update_datetime")

    for table_name, table in schema.get("tables", {}).items():
        resource_type = table.get("resource_type")

        if not resource_type:
            raise ValueError(
                "Each table must specify resource_type. "
                + f"resource_type not found in table {table_name}."
            )

        query_params = table.get("query_params", {})

        # Handle any includes specified in the columns
        for column in table.get("columns", {}).values():
            if "reference_location" in column:
                query_params = _merge_include_query_params_for_location(
                    query_params, column.get("reference_location", "")
                )
        search_string = resource_type
        if query_params is not None and len(query_params) > 0:
            search_string += f"?{urlencode(query_params, True)}"

        count = table.get("results_per_page", count_top)
        since = table.get("earliest_update_datetime", since_top)

        url_dict[table_name] = _generate_search_url(search_string, count, since)

    return url_dict


def _get_reference_directions(schema: dict) -> dict:
    """
    Creates a dictionary mapping indicating how the resources that
    will be used in creating the final output tables relate to each
    other. For any column desired in an output table, it is possible
    for the column to be found in a resource that either a) references
    a given resource, or b) is referenced by the given resource.
    Since each table in the schema is defined with an "anchor" resource
    (the main type of resource determining the number of rows in the
    table), referenced resources of type A can be labeled "backward"
    pointers and referenced resources of type B can be labeled "forward"
    pointers. This mapping is used to efficiently group and aggregate
    related resource data for tabulation.
    :param schema: A user-defined schema, for one or more tables, that
        maps a FHIR resource and element to a specified column in a table.
    :return: A dict containing mappings, for each table, of
      how referenced resources relate to the anchor resource.
    """

    directions_by_table = {}
    for table_name, table_params in schema.get("tables", {}).items():
        anchor_type = table_params.get("resource_type", "")
        directions_by_table[table_name] = {
            "anchor": anchor_type,
            "forward": set(),
            "reverse": {},
        }

        for column_params in table_params.get("columns", {}).values():
            if "reference_location" in column_params:
                [direction, ref_path] = column_params.get(
                    "reference_location", ""
                ).split(":", 1)
                referenced_resource_type = column_params.get("fhir_path", "").split(
                    "."
                )[0]
                if direction == "forward":
                    directions_by_table[table_name][direction].add(
                        referenced_resource_type
                    )
                else:
                    directions_by_table[table_name][direction][
                        referenced_resource_type
                    ] = ref_path

    return directions_by_table


def generate_tables(
    schema_path: pathlib.Path,
    output_params: dict,
    fhir_url: str,
    cred_manager: BaseCredentialManager = None,
) -> None:
    """
    Queries a FHIR server for information, and generates and stores the tables in the
    desired location, according to the supplied schema.

    :param schema_path: A path to the location of a schema config file.
    :param output_params: A dictionary of dictionaries containing the parameters for
        writing each table specified in the schema. For each table in the schema, the
        nested dictionary must contain a directory, filename, and output_type at
        minimum. See `write_data` function for full writing specifications.
    :param fhir_url: A URL to a FHIR server.
    :param cred_manager: The credential manager used to authenticate to the FHIR server.
    """

    # Load schema
    schema = load_schema(schema_path)

    # Load search_urls to query FHIR server
    search_urls = _generate_search_urls(schema=schema)

    for table_name, search_url in search_urls.items():
        pq_writer = None
        next = search_url
        while next is not None:
            # Return set of incremental results and next URL to query
            incremental_results, next = extract_data_from_fhir_search_incremental(
                search_url=urllib.parse.urljoin(fhir_url, next),
                cred_manager=cred_manager,
            )
            # Tabulate data for set of incremental results
            tabulated_incremental_data = tabulate_data(
                incremental_results, schema, table_name
            )

            # Write set of tabulated incremental data
            pq_writer = write_data(
                tabulated_data=tabulated_incremental_data,
                directory=output_params[table_name].get("directory"),
                filename=output_params[table_name].get("filename"),
                output_type=output_params[table_name].get("output_type"),
                db_file=output_params[table_name].get("db_file", None),
                db_tablename=output_params[table_name].get("db_tablename", None),
                pq_writer=pq_writer,
            )
        if pq_writer is not None:
            pq_writer.close()  # pragma: no cover



import copy

from phdi.fhir.geospatial.core import BaseFhirGeocodeClient
from phdi.geospatial.census import CensusGeocodeClient


class CensusFhirGeocodeClient(BaseFhirGeocodeClient):
    """
    Implementation of a geocoding client designed to handle FHIR-
    formatted data using the Census API.
    """

    def __init__(self):
        self.__client = CensusGeocodeClient()

    def geocode_resource(self, resource: dict, overwrite=True) -> dict:
        """
        Performs geocoding on one or more addresses in a given FHIR
        resource and returns either the result or a copy thereof. The original street
        name, number, and any secondary address line information are returned in the
        original form.
        Currently supported resource types are:

            - Patient

        :param resource: The resource whose addresses should be geocoded.
        :param overwrite: Whether to save the geocoding information over
          the raw data, or to create a copy of the given data and write
          over that instead. Defaults to True (write over given data).
        :return: Geocoded resource as a dict.
        """
        # TODO: research additional APIs that return Apt (and other 2nd line addresses)
        # so that address.line can be overwritten
        if not overwrite:
            resource = copy.deepcopy(resource)

        resource_type = resource.get("resourceType", "")
        if resource_type == "Patient":
            self._geocode_patient_resource(resource)

        return resource

    def _geocode_patient_resource(self, patient: dict) -> None:
        """
        Handles geocoding of all addresses in a given patient resource.
        :param patient: The patient resource whose addresses should be geocoded.
        """
        for address in patient.get("address", []):
            address["street"] = " ".join(item for item in address["line"])
            standardized_address = self.__client.geocode_from_dict(address)

            # Update fields with new, standardized information
            if standardized_address:
                address["city"] = standardized_address.city
                address["state"] = standardized_address.state
                address["postalCode"] = standardized_address.postal_code
                self._store_lat_long_extension(
                    address, standardized_address.lat, standardized_address.lng
                )
                self._store_census_tract_extension(
                    address, standardized_address.census_tract
                )

                # Remove dict entry needed only for geocode_from_dict()
                del address["street"]

    def geocode_bundle(self, bundle: dict, overwrite=True) -> dict:
        """
        Performs geocoding on all resources in a given FHIR bundle whose
        resource type is among those supported by the PHDI SDK. Currently,
        this includes:

            - Patient

        :param bundle: A bundle of fhir resources.
        :param overwrite: Whether to overwrite the address data in the given
          bundle's resources (True), or whether to create a copy of the bundle
          and overwrite that instead (False). Defaults to True.
        :return: A FHIR bundle with geocoded address(es).
        """
        if not overwrite:
            bundle = copy.deepcopy(bundle)

        for entry in bundle.get("entry", []):
            self.geocode_resource(entry.get("resource", {}), overwrite=True)

        return bundle
